from typing import List, Tuple, Callable, Union, Iterable
from uuid import uuid4
from pathlib import Path

import numpy as np
import pandas as pd
import dash
from dash_table import DataTable
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output
import webviz_core_components as wcc
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore
from webviz_config import WebvizPluginABC

from .._datainput.inplace_volumes import extract_volumes
from .._abbreviations.volume_terminology import volume_description, volume_unit
from .._abbreviations.number_formatting import table_statistics_base


class InplaceVolumes(WebvizPluginABC):
    """Visualizes inplace volumetric results from
FMU ensembles.

Input can be given either as aggregated `csv` files or as ensemble name(s)
defined in `shared_settings` (with volumetric `csv` files stored per realization).

---

**Using aggregated data**
* **`csvfile`:** Aggregated csvfile with `REAL`, `ENSEMBLE` and `SOURCE` columns \
(absolute path or relative to config file).

**Using data stored per realization**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`volfiles`:**  Key/value pair of csv files E.g. `{geogrid: geogrid--oil.csv}`.
Only relevant if `ensembles` is defined. The key (e.g. `geogrid`) will be used as `SOURCE`.
* **`volfolder`:** Local folder for the `volfiles`.

**Common settings for both input options**
* **`response`:** Optional volume response to visualize initially.

---

?> The input files must follow FMU standards.

* [Example of an aggregated file for `csvfiles`](https://github.com/equinor/\
webviz-subsurface-testdata/blob/master/aggregated_data/volumes.csv).

* [Example of a file per realization that can be used with `ensembles` and `volfiles`]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_history_match/\
realization-0/iter-0/share/results/volumes/geogrid--oil.csv).

**The following columns will be used as available filters, if present:**

* `ZONE`
* `REGION`
* `FACIES`
* `LICENSE`
* `SOURCE` (relevant if calculations are done for multiple grids)


**Remaining columns are seen as volumetric responses.**

All names are allowed (except those mentioned above, in addition to `REAL` and `ENSEMBLE`), \
but the following responses are given more descriptive names automatically:

* `BULK_OIL`: Bulk Volume (Oil)
* `NET_OIL`: Net Volume (Oil)
* `PORE_OIL`: Pore Volume (Oil)
* `HCPV_OIL`: Hydro Carbon Pore Volume (Oil)
* `STOIIP_OIL`: Stock Tank Oil Initially In Place
* `BULK_GAS`: Bulk Volume (Gas)
* `NET_GAS`: Net Volume (Gas)
* `PORV_GAS`: Pore Volume (Gas)
* `HCPV_GAS`: Hydro Carbon Pore Volume (Gas)
* `GIIP_GAS`: Gas Initially In Place
* `RECOVERABLE_OIL`: Recoverable Volume (Oil)
* `RECOVERABLE_GAS`: Recoverable Volume (Gas)

"""

    TABLE_STATISTICS: List[Tuple[str, dict]] = [("Group", {})] + table_statistics_base()

    def __init__(
        self,
        app: dash.Dash,
        csvfile: Path = None,
        ensembles: list = None,
        volfiles: dict = None,
        volfolder: str = "share/results/volumes",
        response: str = "STOIIP_OIL",
    ):

        super().__init__()

        self.csvfile = csvfile if csvfile else None
        if csvfile and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile" or "ensembles" and "volfiles"'
            )
        if csvfile:
            self.volumes: pd.DataFrame = read_csv(csvfile)

        elif ensembles and volfiles:
            self.ens_paths = {
                ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.volfiles = volfiles
            self.volfolder = volfolder
            self.volumes = extract_volumes(
                self.ens_paths, self.volfolder, self.volfiles
            )

        else:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile" or "ensembles" and "volfiles"'
            )

        self.initial_response = response
        self.uid = uuid4()
        self.selectors_id = {x: str(uuid4()) for x in self.selectors}
        self.plotly_theme = app.webviz_settings["theme"].plotly_theme
        self.initial_group: Union[str, None] = None
        if len(self.volumes["ENSEMBLE"].unique()) > 1:
            self.initial_plot = "Box plot"
            self.initial_group = "ENSEMBLE"
        else:
            self.initial_plot = "Per realization"
            self.initial_group = None
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.ids("layout"),
                "content": ("Dashboard displaying in place volumetric results. "),
            },
            {
                "id": self.ids("graph"),
                "content": (
                    "Chart showing results for the current selection. "
                    "Different charts and options can be selected from the menu above."
                ),
            },
            {
                "id": self.ids("table"),
                "content": (
                    "The table shows statistics for the current active selection. "
                    "Rows can be filtered by searching, and sorted by "
                    "clicking on a column header."
                ),
            },
            {
                "id": self.ids("response"),
                "content": "Select the volumetric calculation to display.",
            },
            {
                "id": self.ids("plot-type"),
                "content": (
                    "Controls the type of the visualized chart. "
                    "Per realization shows bars per realization, "
                    "while the boxplot shows the range per sensitivity."
                ),
            },
            {
                "id": self.ids("group"),
                "content": ("Allows grouping of results on a given category."),
            },
            {
                "id": self.ids("filters"),
                "content": (
                    "Filter on different combinations of e.g. zones, facies and regions "
                    "(The options will vary dependent on what was included "
                    "in the calculation.)"
                ),
            },
        ]

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return (
            [(read_csv, [{"csv_file": self.csvfile}])]
            if self.csvfile
            else [
                (
                    extract_volumes,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "volfolder": self.volfolder,
                            "volfiles": self.volfiles,
                        }
                    ],
                )
            ]
        )

    @property
    def vol_columns(self) -> List[str]:
        """List of all columns in dataframe"""
        return list(self.volumes.columns)

    @property
    def all_selectors(self) -> List[str]:
        """List of all possible selectors"""
        return ["SOURCE", "ENSEMBLE", "ZONE", "REGION", "FACIES", "LICENSE"]

    @property
    def plot_types(self) -> List[str]:
        """List of available plots"""
        return ["Histogram", "Per realization", "Box plot"]

    @property
    def selectors(self) -> List[str]:
        """List of available selector columns in dframe"""
        return [x for x in self.all_selectors if x in self.vol_columns]

    @property
    def responses(self) -> List[str]:
        """List of available volume responses in dframe"""
        return [x for x in self.vol_columns if x not in self.selectors and x != "REAL"]

    @property
    def vol_callback_inputs(self) -> List[Input]:
        """Returns all Dash inputs for selecting and filtering volumes
        The number of inputs will vary depending on the available
        selector columns in the volumes dataframe
        """
        inputs = []
        inputs.append(Input(self.ids("response"), "value"))
        inputs.append(Input(self.ids("plot-type"), "value"))
        inputs.append(Input(self.ids("group"), "value"))
        for selector in self.selectors:
            inputs.append(Input(self.selectors_id[selector], "value"))
        return inputs

    @property
    def selector_dropdowns(self) -> List[html.Div]:
        """Makes dropdowns for each selector.
        Args:
            dframe - Volumetrics Dataframe
            selectors - List of selector columns
        Return:
            dcc.Dropdown objects
        """
        dropdowns: List[html.Div] = []
        for selector in self.selectors:
            elements = list(self.volumes[selector].unique())
            if selector in ["ENSEMBLE", "SOURCE"]:
                value = elements[0]
            else:
                value = elements
            dropdowns.append(
                html.Div(
                    children=[
                        html.Details(
                            open=True,
                            children=[
                                html.Summary(selector.lower().capitalize()),
                                wcc.Select(
                                    id=self.selectors_id[selector],
                                    options=[
                                        {"label": i, "value": i} for i in elements
                                    ],
                                    value=value,
                                    multi=True,
                                    size=min(20, len(elements)),
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        )
                    ]
                )
            )
        return dropdowns

    @property
    def plot_options_layout(self) -> wcc.FlexBox:
        """Row layout of dropdowns for plot options"""
        return wcc.FlexBox(
            children=[
                html.Div(
                    children=html.Label(
                        children=[
                            html.Span("Response:", style={"font-weight": "bold"}),
                            dcc.Dropdown(
                                id=self.ids("response"),
                                options=[
                                    {"label": volume_description(i), "value": i}
                                    for i in self.responses
                                ],
                                value=self.initial_response
                                if self.initial_response in self.responses
                                else self.responses[0],
                                clearable=False,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ]
                    )
                ),
                html.Div(
                    children=html.Label(
                        children=[
                            html.Span("Plot type:", style={"font-weight": "bold"}),
                            dcc.Dropdown(
                                id=self.ids("plot-type"),
                                options=[
                                    {"label": i, "value": i} for i in self.plot_types
                                ],
                                value=self.initial_plot,
                                clearable=False,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ]
                    )
                ),
                html.Div(
                    children=html.Label(
                        children=[
                            html.Span("Group by:", style={"font-weight": "bold"}),
                            dcc.Dropdown(
                                id=self.ids("group"),
                                options=[
                                    {"label": i.lower().capitalize(), "value": i}
                                    for i in self.selectors
                                ],
                                value=self.initial_group,
                                placeholder="Not grouped",
                                persistence=True,
                                persistence_type="session",
                            ),
                        ]
                    )
                ),
            ],
        )

    @property
    def layout(self) -> html.Div:
        """Main layout"""
        return html.Div(
            children=[
                wcc.FlexBox(
                    id=self.ids("layout"),
                    children=[
                        html.Div(
                            style={"flex": 1},
                            children=[
                                html.Span("Filters:", style={"font-weight": "bold"}),
                                html.Div(
                                    id=self.ids("filters"),
                                    children=self.selector_dropdowns,
                                ),
                            ],
                        ),
                        html.Div(
                            style={"flex": 5},
                            children=[
                                self.plot_options_layout,
                                html.Div(
                                    style={"height": 400},
                                    children=wcc.Graph(id=self.ids("graph")),
                                ),
                                html.Div(
                                    children=[
                                        html.Div(
                                            id=self.ids("table_title"),
                                            style={"textAlign": "center"},
                                            children="",
                                        ),
                                        DataTable(
                                            id=self.ids("table"),
                                            sort_action="native",
                                            filter_action="native",
                                            page_action="native",
                                            page_size=10,
                                        ),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
            ]
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            [  # type: ignore
                Output(self.ids("graph"), "figure"),
                Output(self.ids("table"), "data"),
                Output(self.ids("table"), "columns"),
                Output(self.ids("table_title"), "children"),
            ],
            self.vol_callback_inputs,
        )
        # TODO(Sigurd) Doesn't seem to make sense with type hints
        # here unless all args are of same type
        def _render_vol_chart(*args):
            """Renders a volume visualization either as a Plotly Graph or
            as a Dash table object.
            The arguments are given by the vol_callback_inputs property
            Args:
                response: The volumetrics response to plot
                plot_type: The type of graph/table to plot
                group: The selector to group the data by
                selections: Active values from the selector columns
            Return:
                Plotly Graph/dash_table.DataTable
            """
            # pylint: disable=line-too-long
            # TODO(Sigurd) Tricky to figure out the type hints without a bit of guesswork here due to the use of *args
            # The reasoning below is:
            #   response:   ids("response")  dcc.Dropdown(multi=False, clearable=False) => str
            #   plot_type:  ids("plot_type") dcc.Dropdown(multi=False, clearable=False) => str
            #   group:      ids("group")     dcc.Dropdown(multi=False, clearable=True)  => Union[str, None]
            #   selection:  multiple ids     n*wcc.Select(multi=?, clearable=?)         => Tuple[Union[str, List[str], None], ...]
            response: str = args[0]
            plot_type: str = args[1]
            group: Union[str, None] = args[2]
            selections: Tuple[Union[str, List[str], None], ...] = args[3:]
            data = self.volumes
            data = filter_dataframe(data, self.selectors, selections)

            # If not grouped make one trace
            if not group:
                dframe = data.groupby("REAL").sum().reset_index()
                plot_traces = [plot_data(plot_type, dframe, response, "Total")]
                table = [plot_table(dframe, response, "Total")]
            # Else make one trace for each group member
            else:
                plot_traces = []
                table = []
                for name, vol_group_df in data.groupby(group):
                    dframe = vol_group_df.groupby("REAL").sum().reset_index()
                    trace = plot_data(plot_type, dframe, response, name)
                    if trace is not None:
                        plot_traces.append(trace)
                        table.append(plot_table(dframe, response, name))
            # Column specification
            columns = [
                {**{"name": i[0], "id": i[0]}, **i[1]}
                for i in InplaceVolumes.TABLE_STATISTICS
            ]
            # Make a graph object and return
            return (
                {
                    "data": plot_traces,
                    "layout": plot_layout(plot_type, response, theme=self.plotly_theme),
                },
                table,
                columns,
                f"{volume_description(response)} [{volume_unit(response)}]",
            )

        @app.callback(
            [
                Output(self.selectors_id["ENSEMBLE"], "multi"),
                Output(self.selectors_id["ENSEMBLE"], "value"),
                Output(self.selectors_id["ENSEMBLE"], "size"),
            ],
            [Input(self.ids("group"), "value")],
        )
        def _set_iteration_selector(
            group_by: Union[str, None]
        ) -> Tuple[bool, Union[str, list], int]:
            """If iteration is selected as group by set the iteration
            selector to allow multiple selections, else use single selection
            """
            selectors = list(self.volumes["ENSEMBLE"].unique())
            if group_by == "ENSEMBLE":
                return True, selectors, len(selectors)

            return False, selectors[0], 1

        if "SOURCE" in self.selectors:

            @app.callback(
                [
                    Output(self.selectors_id["SOURCE"], "multi"),
                    Output(self.selectors_id["SOURCE"], "value"),
                    Output(self.selectors_id["SOURCE"], "size"),
                ],
                [Input(self.ids("group"), "value")],
            )
            def _set_source_selector(
                group_by: Union[str, None]
            ) -> Tuple[bool, Union[str, list], int]:
                """If iteration is selected as group by set the iteration
                selector to allow multiple selections, else use single selection
                """
                selectors = list(self.volumes["SOURCE"].unique())
                if group_by == "SOURCE" and "SOURCE" in self.selectors:
                    return True, selectors, len(selectors)

                return False, selectors[0], 1


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def plot_data(
    plot_type: str, dframe: pd.DataFrame, response: str, name: str
) -> Union[dict, None]:
    values = dframe[response]

    output: Union[dict, None] = None
    if plot_type == "Histogram":
        if values.nunique() == 1:
            values = values[0]
        output = {"x": values, "type": "histogram", "name": name}
    elif plot_type == "Box plot":
        output = {"y": values, "name": name, "type": "box"}
    elif plot_type == "Per realization":
        output = {"y": values, "x": dframe["REAL"], "name": name, "type": "bar"}
    else:
        output = None

    return output


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def plot_table(dframe: pd.DataFrame, response: str, name: str) -> Union[dict, None]:
    values = dframe[response]

    output: Union[dict, None] = None
    try:
        output = {
            "Group": str(name),
            "Minimum": values.min(),
            "Maximum": values.max(),
            "Mean": values.mean(),
            "Stddev": values.std(),
            "P10": np.percentile(values, 90),
            "P90": np.percentile(values, 10),
        }
    except KeyError:
        output = None

    return output


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def plot_layout(plot_type: str, response: str, theme: dict) -> dict:
    layout = {}
    layout.update(theme["layout"])
    layout["height"] = 400
    if plot_type == "Histogram":
        layout.update(
            {
                "barmode": "overlay",
                "bargap": 0.01,
                "bargroupgap": 0.2,
                "xaxis": {
                    "title": f"{volume_description(response)} [{volume_unit(response)}]"
                },
                "yaxis": {"title": "Count"},
            }
        )
    elif plot_type == "Box plot":
        layout.update(
            {
                "yaxis": {
                    "title": f"{volume_description(response)} [{volume_unit(response)}]"
                }
            }
        )
    else:
        layout.update(
            {
                "margin": {"l": 60, "r": 40, "b": 30, "t": 10},
                "yaxis": {
                    "title": f"{volume_description(response)} [{volume_unit(response)}]"
                },
                "xaxis": {"title": "Realization"},
            }
        )

    # output["colorway"] = colors
    return layout


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_dataframe(
    dframe: pd.DataFrame,
    columns: Union[str, List[str]],
    column_values: Iterable[Union[str, List[str]]],
) -> pd.DataFrame:
    df = dframe.copy()
    if not isinstance(columns, list):
        columns = [columns]
    for filt, col in zip(column_values, columns):
        if isinstance(filt, list):
            df = df.loc[df[col].isin(filt)]
        else:
            df = df.loc[df[col] == filt]
    return df


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)



import math
import json
import pathlib
import warnings
from typing import Optional, Union, List, Tuple


_DATA_PATH = pathlib.Path(__file__).parent.absolute() / "abbreviation_data"

SI_PREFIXES = json.loads((_DATA_PATH / "si_prefixes.json").read_text())


def table_statistics_base() -> List[Tuple[str, dict]]:
    return [
        (
            i,
            {
                "type": "numeric",
                "format": {
                    "locale": {"symbol": ["", ""]},
                    "specifier": "$.4s",
                },
            },
        )
        if i != "Stddev"
        else (
            i,
            {
                "type": "numeric",
                "format": {
                    "locale": {"symbol": ["", ""]},
                    "specifier": "$.3s",
                },
            },
        )
        for i in ["Mean", "Stddev", "Minimum", "P90", "P10", "Maximum"]
    ]


def si_prefixed(
    number: float,
    number_format: str = "",
    unit: str = "",
    spaced: bool = True,
    locked_si_prefix: Optional[Union[str, int]] = None,
) -> str:
    """
    Formats a float as a string with SI-prefix and optionally a unit. Useful when you cannot/
    don't want to use the Dash d3-based formatting.
    Arguments:
    * `number` (float): Number to format
    * `number_format` (str): Format of the numeric part based on the Python Format Specification
    Mini-Language e.g. '.3g' for 3 significant digits, '.2f' for two decimals, or '.0f' for no
    decimals.
    * `unit` (str): String to append at the end as a unit
    * `spaced` (bool): Include a space between last numerical digit and SI-prefix.
    * `locked_si_prefix` (str or int): Lock the SI prefix to either a string (e.g. 'm' (milli) or
     'M' (mega)) or an integer which is the base 10 exponent (e.g. 3 for kilo, -3 for milli).
    """

    def number_formatter(number_base: float, si_prefix: str) -> str:
        return (
            f"{number_base:{number_format}} {si_prefix}{unit}"
            if spaced
            else f"{number_base:{number_format}}{si_prefix}{unit}"
        )

    if locked_si_prefix is not None:
        # Make sure locked_si_prefix is a string to avoid == issues
        locked_si_prefix = str(locked_si_prefix)

        if str(locked_si_prefix) == "" or locked_si_prefix == str(0):
            return number_formatter(number, "")

        for key, value in SI_PREFIXES.items():
            if locked_si_prefix == str(value) or locked_si_prefix == str(key):
                return number_formatter(number / (10 ** float(key)), value)
        # If we get further, an invalid si_prefix has been inputed, will throw warning, but
        # will continue and use non-locked prefix.
        warnings.warn(
            (
                f"An invalid locked_si_prefix={locked_si_prefix} was used. Ignoring the locked "
                "prefix and calculate instead."
            ),
            UserWarning,
        )
    # Calculate prefix based on value when not locked
    # Zero is a special case
    if number == 0:
        return number_formatter(0, SI_PREFIXES["0"])

    (exp_div_3, log10_rem) = divmod(math.log10(math.fabs(number)), 3)
    # Take log 10 and then mod 3 as we have one prefix per 10^3, the divisor*3 is then the exponent
    return (
        number_formatter(-(10 ** log10_rem), SI_PREFIXES[str(int(exp_div_3 * 3))])
        if (number < 0)
        else number_formatter(10 ** log10_rem, SI_PREFIXES[str(int(exp_div_3 * 3))])
    )



########################################
#
#  Copyright (C) 2020-     Equinor ASA
#
########################################

from typing import Callable, Dict, List, Tuple, Union, Any

import pandas as pd
import dash
import dash_html_components as html
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_core_components as dcc
import webviz_core_components as wcc
from webviz_config.common_cache import CACHE
from webviz_config import WebvizPluginABC

from .._datainput.pvt_data import load_pvt_dataframe, load_pvt_csv


class PvtPlot(WebvizPluginABC):
    """Visualizes formation volume factor and viscosity data \
    for oil, gas and water from both **csv**, Eclipse **init** and **include** files.

    !> The plugin supports variations in PVT between ensembles, but not between \
    realizations in the same ensemble.
    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`pvt_relative_file_path`:** Local path to a csv file in each \
        realization with dumped pvt data.
    * **`read_from_init_file`:** A boolean flag stating if data shall be \
        read from an Eclipse INIT file instead of an INCLUDE file. \
        This is only used when **pvt_relative_file_path** is not given.
    * **`drop_ensemble_duplicates`:** A boolean flag stating if ensembles \
        which are holding duplicate data of other ensembles shall be dropped. \
        Defaults to False.

    ---
    The minimum requirement is to define `ensembles`.

    If no `pvt_relative_file_path` is given, the PVT data will be extracted automatically
    from the simulation decks of individual realizations using `fmu_ensemble` and `ecl2df`.
    If the `read_from_init_file` flag is set to True, the extraction procedure in
    `ecl2df` will be replaced by an individual extracting procedure that reads the
    normalized Eclipse INIT file.
    Note that the latter two extraction methods can be very slow for larger data and are therefore
    not recommended unless you have a very simple model/data deck.
    If the `drop_ensemble_duplicates` flag is set to True, any ensembles which are holding
    duplicate data of other ensembles will be dropped.

    `pvt_relative_file_path` is a path to a file stored per realization (e.g. in \
    `share/results/tables/pvt.csv`). `pvt_relative_file_path` columns:
    * One column named `KEYWORD` or `TYPE`: with Flow/Eclipse style keywords
        (e.g. `PVTO` and `PVDG`).
    * One column named `PVTNUM` with integer `PVTNUM` regions.
    * One column named `RATIO` or `R` with the gas-oil-ratio as the primary variate.
    * One column named `PRESSURE` with the fluids pressure as the secondary variate.
    * One column named `VOLUMEFACTOR` as the first covariate.
    * One column named `VISCOSITY` as the second covariate.

    The file can e.g. be dumped to disc per realization by a forward model in ERT using
    `ecl2df`.
    """

    PHASES = ["OIL", "GAS", "WATER"]

    def __init__(
        self,
        app: dash.Dash,
        ensembles: List[str],
        pvt_relative_file_path: str = None,
        read_from_init_file: bool = False,
        drop_ensemble_duplicates: bool = False,
    ):

        super().__init__()

        self.ensemble_paths = {
            ensemble: app.webviz_settings["shared_settings"]["scratch_ensembles"][
                ensemble
            ]
            for ensemble in ensembles
        }

        self.plotly_theme = app.webviz_settings["theme"].plotly_theme

        self.pvt_relative_file_path = pvt_relative_file_path

        self.read_from_init_file = read_from_init_file

        self.drop_ensemble_duplicates = drop_ensemble_duplicates

        if self.pvt_relative_file_path is None:
            self.pvt_data_frame = load_pvt_dataframe(
                self.ensemble_paths,
                use_init_file=read_from_init_file,
                drop_ensemble_duplicates=drop_ensemble_duplicates,
            )
        else:
            # Load data from all ensembles into a pandas DataFrame
            self.pvt_data_frame = load_pvt_csv(
                ensemble_paths=self.ensemble_paths,
                csv_file=self.pvt_relative_file_path,
                drop_ensemble_duplicates=drop_ensemble_duplicates,
            )

            self.pvt_data_frame = self.pvt_data_frame.rename(
                str.upper, axis="columns"
            ).rename(
                columns={"TYPE": "KEYWORD", "RS": "RATIO", "R": "RATIO", "GOR": "RATIO"}
            )

            # Ensure that the identifier string "KEYWORD" is contained in the header columns
            if "KEYWORD" not in self.pvt_data_frame.columns:
                raise ValueError(
                    "There has to be a KEYWORD or TYPE column with corresponding Eclipse keyword."
                )

        self.phases_additional_info: List[str] = []
        if self.pvt_data_frame["KEYWORD"].str.contains("PVTO").any():
            self.phases_additional_info.append("PVTO")
        elif self.pvt_data_frame["KEYWORD"].str.contains("PVDO").any():
            self.phases_additional_info.append("PVDO")
        if self.pvt_data_frame["KEYWORD"].str.contains("PVTG").any():
            self.phases_additional_info.append("PVTG")
        elif self.pvt_data_frame["KEYWORD"].str.contains("PVDG").any():
            self.phases_additional_info.append("PVDG")
        if self.pvt_data_frame["KEYWORD"].str.contains("PVTW").any():
            self.phases_additional_info.append("PVTW")

        self.set_callbacks(app)

    @property
    def phases(self) -> Dict[str, str]:
        phase_descriptions: Dict[str, str] = {}
        for i in range(0, len(PvtPlot.PHASES)):
            phase_descriptions[PvtPlot.PHASES[i]] = self.phases_additional_info[i]
        return phase_descriptions

    @property
    def ensembles(self) -> List[str]:
        return list(self.pvt_data_frame["ENSEMBLE"].unique())

    @property
    def pvtnums(self) -> List[str]:
        return list(self.pvt_data_frame["PVTNUM"].unique())

    @property
    def ensemble_colors(self) -> Dict[str, List[str]]:
        return {
            ensemble: self.plotly_theme["layout"]["colorway"][
                self.ensembles.index(ensemble)
            ]
            for ensemble in self.ensembles
        }

    @property
    def pvtnum_colors(self) -> Dict[str, List[str]]:
        return {
            pvtnum: self.plotly_theme["layout"]["colorway"][self.pvtnums.index(pvtnum)]
            for pvtnum in self.pvtnums
        }

    @property
    def color_options(self) -> List[str]:
        """Options to color by"""
        return ["ENSEMBLE", "PVTNUM"]

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": "Dashboard displaying formation volume factor and viscosity"
                " data of either Oil, Gas or Water",
            },
            {
                "id": self.uuid("graph"),
                "content": (
                    "Visualization of curves. "
                    "Different options can be set in the menu to the left."
                    "You can also toggle data on/off by clicking at the legend."
                ),
            },
            {
                "id": self.uuid("color_by_selector"),
                "content": ("Choose the basis for your colormap."),
            },
            {
                "id": self.uuid("ensemble_selector"),
                "content": ("Select ensembles."),
            },
            {
                "id": self.uuid("phase_selector"),
                "content": (
                    "Choose a phase. Formation volume factor and viscosity data will be"
                    " shown for the selected phase in separate plots."
                ),
            },
            {
                "id": self.uuid("pvtnum_selector"),
                "content": ("Choose PVTNUM regions."),
            },
        ]

    @staticmethod
    def set_grid_layout(columns: int, padding: int = 0) -> Dict[str, str]:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
            "padding": f"{padding}px",
        }

    @property
    def layout(self) -> wcc.FlexBox:
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                html.Div(
                    id=self.uuid("filters"),
                    style={"flex": "1"},
                    children=[
                        html.Label(
                            id=self.uuid("color_by_selector"),
                            children=[
                                html.Span("Color by:", style={"font-weight": "bold"}),
                                dcc.Dropdown(
                                    id=self.uuid("color_by"),
                                    clearable=False,
                                    options=[
                                        {
                                            "label": i.lower().capitalize(),
                                            "value": i,
                                        }
                                        for i in self.color_options
                                    ],
                                    value=self.color_options[0],
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                        html.Label(
                            id=self.uuid("ensemble_selector"),
                            children=[
                                html.Span("Ensembles:", style={"font-weight": "bold"}),
                                dcc.Dropdown(
                                    id=self.uuid("ensemble"),
                                    clearable=False,
                                    multi=True,
                                    options=[
                                        {"label": i, "value": i} for i in self.ensembles
                                    ],
                                    value=self.ensembles,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                        html.Label(
                            id=self.uuid("phase_selector"),
                            children=[
                                html.Span(
                                    "Phase:",
                                    style={"font-weight": "bold"},
                                ),
                                dcc.Dropdown(
                                    id=self.uuid("phase"),
                                    clearable=False,
                                    options=[
                                        {
                                            "label": f"{value.lower().capitalize()} ({info})",
                                            "value": value,
                                        }
                                        for value, info in self.phases.items()
                                    ],
                                    multi=False,
                                    value=list(self.phases.keys())[0],
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                        html.Label(
                            id=self.uuid("pvtnum_selector"),
                            children=[
                                html.Span("Pvtnum:", style={"font-weight": "bold"}),
                                dcc.Dropdown(
                                    id=self.uuid("pvtnum"),
                                    clearable=False,
                                    multi=False,
                                    options=[
                                        {"label": i, "value": i} for i in self.pvtnums
                                    ],
                                    value=self.pvtnums[0],
                                    persistence=True,
                                    persistence_type="session",
                                ),
                            ],
                        ),
                    ],
                ),
                html.Div(
                    id=self.uuid("graphs"),
                    style={"flex": "4"},
                    children=wcc.Graph(id=self.uuid("graph")),
                ),
                dcc.Store(
                    id=self.uuid("init_callback"), storage_type="session", data=True
                ),
            ],
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            [
                Output(self.uuid("graph"), "figure"),
                Output(self.uuid("init_callback"), "data"),
            ],
            [
                Input(self.uuid("phase"), "value"),
                Input(self.uuid("ensemble"), "value"),
                Input(self.uuid("pvtnum"), "value"),
                Input(self.uuid("color_by"), "value"),
            ],
            [State(self.uuid("init_callback"), "data")],
        )
        def _update_graph(
            phase: str,
            ensembles: Union[List[str], str],
            pvtnums: Union[List[str], str],
            color_by: str,
            init_callback: bool,
        ) -> List[Union[Dict[str, Union[dict, List[dict]]], bool]]:
            if (
                not init_callback
                and not dash.callback_context.triggered[0]["prop_id"].split(".")[0]
            ):
                raise PreventUpdate
            if not phase:
                raise PreventUpdate
            if ensembles is None:
                ensembles = []
            if not isinstance(ensembles, list):
                ensembles = [ensembles]
            if not isinstance(pvtnums, list):
                pvtnums = [pvtnums]

            data_frame = filter_data_frame(self.pvt_data_frame, ensembles, pvtnums)

            if color_by == "ENSEMBLE":
                colors = self.ensemble_colors
            elif color_by == "PVTNUM":
                colors = self.pvtnum_colors

            layout = plot_layout(phase, color_by, self.plotly_theme["layout"])

            traces = add_realization_traces(data_frame, color_by, colors, phase)

            return [{"data": traces, "layout": layout}, False]

        @app.callback(
            [
                Output(self.uuid("ensemble"), "multi"),
                Output(self.uuid("ensemble"), "value"),
            ],
            [Input(self.uuid("color_by"), "value")],
        )
        def _set_ensemble_selector(color_by: str) -> Tuple[bool, Union[str, List[str]]]:
            # pylint: disable=unused-argument
            """If ensemble is selected as color by, set the ensemble
            selector to allow multiple selections
            """
            if (
                dash.callback_context.triggered is None
                or not dash.callback_context.triggered[0]["prop_id"].split(".")[0]
            ):
                raise PreventUpdate

            if color_by == "ENSEMBLE":
                return True, self.ensembles

            # Note: Reimplement using stored_ensemble as soon as it is working properly.
            return (
                False,
                self.ensembles[0],
            )

        @app.callback(
            [
                Output(self.uuid("pvtnum"), "multi"),
                Output(self.uuid("pvtnum"), "value"),
            ],
            [Input(self.uuid("color_by"), "value")],
        )
        # pylint:
        def _set_pvtnum_selector(color_by: str) -> Tuple[bool, Union[str, List[str]]]:
            # pylint: disable=unused-argument
            """If pvtnum is selected as color by, set the pvtnum
            selector to allow multiple selections
            """
            if (
                dash.callback_context.triggered is None
                or not dash.callback_context.triggered[0]["prop_id"].split(".")[0]
            ):
                raise PreventUpdate

            if color_by == "PVTNUM":
                return True, self.pvtnums

            # Note: Reimplement using stored_pvtnum as soon as it is working properly.
            return (
                False,
                self.pvtnums[0],
            )

    def add_webvizstore(
        self,
    ) -> List[Tuple[Callable, List[Dict[str, Any]]]]:
        return (
            [
                (
                    load_pvt_dataframe,
                    [
                        {
                            "ensemble_paths": self.ensemble_paths,
                            "use_init_file": self.read_from_init_file,
                            "drop_ensemble_duplicates": self.drop_ensemble_duplicates,
                        }
                    ],
                )
            ]
            if self.pvt_relative_file_path is None
            else [
                (
                    load_pvt_csv,
                    [
                        {
                            "ensemble_paths": self.ensemble_paths,
                            "csv_file": self.pvt_relative_file_path,
                            "drop_ensemble_duplicates": self.drop_ensemble_duplicates,
                        }
                    ],
                )
            ]
        )


def filter_data_frame(
    data_frame: pd.DataFrame, ensembles: List[str], pvtnums: List[str]
) -> pd.DataFrame:

    data_frame = data_frame.copy()
    data_frame = data_frame.loc[data_frame["ENSEMBLE"].isin(ensembles)]
    data_frame = data_frame.loc[data_frame["PVTNUM"].isin(pvtnums)]
    return data_frame.fillna(0)


def add_realization_traces(
    data_frame: pd.DataFrame, color_by: str, colors: Dict[str, List[str]], phase: str
) -> List[dict]:
    """Renders line traces for individual realizations"""
    # pylint: disable-msg=too-many-locals

    traces = []

    dim1_column_name = "RATIO"

    if phase == "OIL":
        data_frame = data_frame.loc[
            (data_frame["KEYWORD"] == "PVTO") | (data_frame["KEYWORD"] == "PVDO")
        ]
    elif phase == "GAS":
        data_frame = data_frame.loc[
            (data_frame["KEYWORD"] == "PVTG") | (data_frame["KEYWORD"] == "PVDG")
        ]
        dim1_column_name = "PRESSURE"
    else:
        data_frame = data_frame.loc[data_frame["KEYWORD"] == "PVTW"]

    border_value_pressure: Dict[str, list] = {}
    border_value_viscosity: Dict[str, list] = {}
    border_value_volumefactor: Dict[str, list] = {}
    constant_group = (
        data_frame["PVTNUM"].iloc[0]
        if color_by == "ENSEMBLE"
        else data_frame["ENSEMBLE"].iloc[0]
    )

    for (group, grouped_data_frame) in data_frame.groupby(color_by):
        for set_no, set_value in enumerate(
            grouped_data_frame[dim1_column_name].unique()
        ):
            for realization_no, (realization, realization_data_frame) in enumerate(
                grouped_data_frame.groupby("REAL")
            ):
                if group not in border_value_pressure:
                    border_value_pressure[group] = []
                    border_value_viscosity[group] = []
                    border_value_volumefactor[group] = []
                try:
                    border_value_pressure[group].append(
                        realization_data_frame.loc[
                            realization_data_frame[dim1_column_name] == set_value
                        ]["PRESSURE"].iloc[0]
                    )
                    border_value_volumefactor[group].append(
                        realization_data_frame.loc[
                            realization_data_frame[dim1_column_name] == set_value
                        ]["VOLUMEFACTOR"].iloc[0]
                    )
                    border_value_viscosity[group].append(
                        realization_data_frame.loc[
                            realization_data_frame[dim1_column_name] == set_value
                        ]["VISCOSITY"].iloc[0]
                    )
                except IndexError as exc:
                    raise IndexError(
                        "This error is most likely due to PVT differences between "
                        "realizations within the same ensemble. This is currently not "
                        "supported."
                    ) from exc

                hovertext: Union[str, list] = ""
                if phase == "OIL":
                    hovertext = (
                        "{} Pvtnum: {}<br />Realization: {}, Ensemble: {}".format(
                            f"Rs = {set_value}"
                            if realization_data_frame["KEYWORD"]
                            .str.contains("PVTO")
                            .any()
                            else "",
                            group if color_by == "PVTNUM" else constant_group,
                            realization,
                            group if color_by == "ENSEMBLE" else constant_group,
                        )
                    )
                elif phase == "GAS":
                    hovertext = [
                        "{}"
                        "Pvtnum: "
                        "{}<br>"
                        "Realization: {}, Ensemble: "
                        "{}".format(
                            "Rv = {}, ".format(
                                realization_data_frame.loc[
                                    (realization_data_frame["PRESSURE"] == y)
                                    & (realization_data_frame["VOLUMEFACTOR"] == x)
                                ]["RATIO"].iloc[0]
                            )
                            if realization_data_frame["KEYWORD"]
                            .str.contains("PVTG")
                            .any()
                            else "",
                            group if color_by == "PVTNUM" else constant_group,
                            realization,
                            group if color_by == "ENSEMBLE" else constant_group,
                        )
                        for x, y in zip(
                            realization_data_frame.loc[
                                realization_data_frame["PRESSURE"] == set_value
                            ].VOLUMEFACTOR,
                            realization_data_frame.loc[
                                realization_data_frame["PRESSURE"] == set_value
                            ].PRESSURE,
                        )
                    ]
                else:
                    hovertext = (
                        f"Pvtnum: {group if color_by == 'PVTNUM' else constant_group}<br />"
                        f"Realization: {realization}, "
                        f"Ensemble: {group if color_by == 'ENSEMBLE' else constant_group}"
                    )

                traces.extend(
                    [
                        {
                            "type": "scatter",
                            "x": realization_data_frame.loc[
                                realization_data_frame[dim1_column_name] == set_value
                            ]["PRESSURE"],
                            "y": realization_data_frame.loc[
                                realization_data_frame[dim1_column_name] == set_value
                            ]["VOLUMEFACTOR"],
                            "xaxis": "x",
                            "yaxis": "y",
                            "hovertext": hovertext,
                            "name": group,
                            "legendgroup": group,
                            "marker": {
                                "color": colors.get(
                                    group, colors[list(colors.keys())[-1]]
                                )
                            },
                            "showlegend": realization_no == 0 and set_no == 0,
                        }
                    ]
                )

                if phase == "GAS":
                    hovertext = [
                        "{}Pvtnum: {}<br>Realization: {}, Ensemble: {}".format(
                            "Rv = {}, ".format(
                                realization_data_frame.loc[
                                    (realization_data_frame["PRESSURE"] == y)
                                    & (realization_data_frame["VISCOSITY"] == x)
                                ]["RATIO"].iloc[0]
                            )
                            if realization_data_frame["KEYWORD"]
                            .str.contains("PVTG")
                            .any()
                            else "",
                            group if color_by == "PVTNUM" else constant_group,
                            realization,
                            group if color_by == "ENSEMBLE" else constant_group,
                        )
                        for x, y in zip(
                            realization_data_frame.loc[
                                realization_data_frame["PRESSURE"] == set_value
                            ].VISCOSITY,
                            realization_data_frame.loc[
                                realization_data_frame["PRESSURE"] == set_value
                            ].PRESSURE,
                        )
                    ]

                traces.extend(
                    [
                        {
                            "type": "scatter",
                            "x": realization_data_frame.loc[
                                realization_data_frame[dim1_column_name] == set_value
                            ]["PRESSURE"],
                            "y": realization_data_frame.loc[
                                realization_data_frame[dim1_column_name] == set_value
                            ]["VISCOSITY"],
                            "xaxis": "x2",
                            "yaxis": "y2",
                            "hovertext": hovertext,
                            "name": group,
                            "legendgroup": group,
                            "marker": {
                                "color": colors.get(
                                    group, colors[list(colors.keys())[-1]]
                                )
                            },
                            "showlegend": False,
                        }
                    ]
                )

    for group in border_value_pressure:
        traces.extend(
            [
                {
                    "type": "scatter",
                    "mode": "lines",
                    "x": border_value_pressure[group],
                    "y": border_value_volumefactor[group],
                    "xaxis": "x",
                    "yaxis": "y",
                    "line": {
                        "width": 1,
                        "color": colors.get(group, colors[list(colors.keys())[-1]]),
                    },
                    "showlegend": False,
                }
            ]
        )
        traces.extend(
            [
                {
                    "type": "scatter",
                    "mode": "lines",
                    "x": border_value_pressure[group],
                    "y": border_value_viscosity[group],
                    "xaxis": "x2",
                    "yaxis": "y2",
                    "line": {
                        "width": 1,
                        "color": colors.get(group, colors[list(colors.keys())[-1]]),
                    },
                    "showlegend": False,
                }
            ]
        )
    return traces


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def plot_layout(phase: str, color_by: str, theme: dict) -> dict:
    """
    Constructing plot layout from scratch as it is more responsive than plotly subplots package.
    """
    titles = [
        "{} Formation Volume Factor".format(phase.lower().capitalize()),
        "{} Viscosity".format(phase.lower().capitalize()),
    ]
    layout = {}
    layout.update(theme)
    layout.update({"hovermode": "closest"})
    # create subplots
    layout.update(
        {
            "annotations": [
                {
                    "showarrow": False,
                    "text": titles[0],
                    "x": 0.5,
                    "xanchor": "center",
                    "xref": "paper",
                    "y": 1.0,
                    "yanchor": "bottom",
                    "yref": "paper",
                    "font": {"size": 16},
                },
                {
                    "showarrow": False,
                    "text": titles[1],
                    "x": 0.5,
                    "xanchor": "center",
                    "xref": "paper",
                    "y": 0.475,
                    "yanchor": "bottom",
                    "yref": "paper",
                    "font": {"size": 16},
                },
            ],
        }
    )

    layout["legend"] = {"title": {"text": color_by.lower().capitalize()}}
    # format axes
    layout.update(
        {
            "xaxis": {
                "automargin": True,
                "zeroline": False,
                "anchor": "y",
                "domain": [0.0, 1.0],
                "matches": "x2",
                "showticklabels": False,
                "showgrid": True,
            },
            "xaxis2": {
                "automargin": True,
                "ticks": "",
                "showticklabels": True,
                "zeroline": False,
                "anchor": "y2",
                "domain": [0.0, 1.0],
                "title": {"text": "Pressure [barsa]", "standoff": 15},
                "showgrid": True,
            },
            "yaxis": {
                "automargin": True,
                "ticks": "",
                "zeroline": False,
                "anchor": "x",
                "domain": [0.525, 1.0],
                "title": {
                    "text": "{} Formation Volume Factor [rm3/sm3]".format(
                        phase.lower().capitalize()
                    )
                },
                "type": "linear",
                "showgrid": True,
            },
            "yaxis2": {
                "automargin": True,
                "ticks": "",
                "zeroline": False,
                "anchor": "x2",
                "domain": [0.0, 0.475],
                "title": {
                    "text": "{} Viscosity [cP]".format(phase.lower().capitalize())
                },
                "type": "linear",
                "showgrid": True,
            },
            "height": 800,
            "margin": {"t": 20, "b": 0},
            "plot_bgcolor": "rgba(0,0,0,0)",
        }
    )
    return layout



from typing import Optional

import xtgeo
from webviz_config.common_cache import CACHE


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def load_grid(gridpath: str) -> xtgeo.Grid:
    return xtgeo.grid_from_file(gridpath)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def load_grid_parameter(
    grid: Optional[xtgeo.Grid], gridparameterpath: str
) -> xtgeo.GridProperty:
    return xtgeo.gridproperty_from_file(gridparameterpath, grid=grid)



from typing import Callable, List, Tuple, Dict
import json
from uuid import uuid4
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.stats import chi2
import dash
import dash_html_components as html
import webviz_subsurface_components as wsc
from webviz_config import WebvizPluginABC

from .._datainput.history_match import extract_mismatch


class HistoryMatch(WebvizPluginABC):
    """Visualizes the quality of the history match.

---

* **`ensembles`:** List of the ensembles in `shared_settings` to visualize.
* **`observation_file`:** Path to the observation `.yaml` file \
(absolute or relative to config file).

---
Parameter values are extracted automatically from the `parameters.txt` files
of the individual realizations of your given `ensembles`, using the `fmu-ensemble` library.

?> The `observation_file` is a common (optional) file for all ensembles, which currently has to \
be made manually. [An example of the format can be found here]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_history_match/share/\
observations/observations.yml).
"""

    def __init__(self, app: dash.Dash, ensembles: List[str], observation_file: Path):

        super().__init__()

        self.observation_file = observation_file

        self.ensembles = ensembles
        self.ens_paths = {
            ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][ens]
            for ens in ensembles
        }

        data = extract_mismatch(self.ens_paths, self.observation_file)
        self.hm_data = json.dumps(self._prepare_data(data))

        self.hm_id = "hm-id-{}".format(uuid4())

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (
                extract_mismatch,
                [
                    {
                        "ens_paths": self.ens_paths,
                        "observation_file": self.observation_file,
                    }
                ],
            )
        ]

    def _prepare_data(self, data: pd.DataFrame) -> dict:
        data = data.copy().reset_index()

        num_obs_groups = len(data.obs_group_name.unique())

        data["avg_pos"] = data["total_pos"] / data["number_data_points"]
        data["avg_neg"] = data["total_neg"] / data["number_data_points"]

        iterations = []
        for ensemble in self.ensembles:
            df = data[data.ensemble_name == ensemble]
            iterations.append(df.groupby("obs_group_name").mean())

        sorted_iterations = HistoryMatch._sort_iterations(iterations)

        iterations_dict = HistoryMatch._iterations_to_dict(
            sorted_iterations, self.ensembles
        )

        confidence_sorted = _get_sorted_edges(num_obs_groups)
        confidence_unsorted = _get_unsorted_edges()

        data = {}
        data["iterations"] = iterations_dict
        data["confidence_interval_sorted"] = confidence_sorted
        data["confidence_interval_unsorted"] = confidence_unsorted

        return data

    @staticmethod
    def _sort_iterations(iterations: List[pd.DataFrame]) -> List[pd.DataFrame]:
        sorted_data = []

        for df in iterations:
            sorted_df = df.copy()

            sorted_data.append(
                sorted_df.assign(f=sorted_df["avg_pos"] + sorted_df["avg_neg"])
                .sort_values("f", ascending=False)
                .drop("f", axis=1)
            )

        return sorted_data

    @staticmethod
    def _iterations_to_dict(
        iterations: List[pd.DataFrame], labels: List[str]
    ) -> List[dict]:
        retval = []

        for iteration, label in zip(iterations, labels):
            retval.append(
                {
                    "name": label,
                    "positive": iteration["avg_pos"].tolist(),
                    "negative": iteration["avg_neg"].tolist(),
                    "labels": iteration.index.tolist(),
                }
            )

        return retval

    @property
    def layout(self) -> html.Div:
        return html.Div([wsc.HistoryMatch(id=self.hm_id, data=self.hm_data)])


def _get_unsorted_edges() -> dict:
    """P10 - P90 unsorted edge coordinates"""

    retval = {"low": chi2.ppf(0.1, 1), "high": chi2.ppf(0.9, 1)}

    return retval


def _get_sorted_edges(number_observation_groups: int) -> Dict[str, list]:
    """P10 - P90 sorted edge coordinates"""

    monte_carlo_iterations = 100000

    sorted_values = np.empty((number_observation_groups, monte_carlo_iterations))

    for i in range(monte_carlo_iterations):
        sorted_values[:, i] = np.sort(
            np.random.chisquare(df=1, size=number_observation_groups)
        )

    sorted_values = np.flip(sorted_values, 0)

    p10 = np.percentile(sorted_values, 90, axis=1)
    p90 = np.percentile(sorted_values, 10, axis=1)

    # Dictionary with two arrays (P10, P90). Each array of length equal
    # to number of observation groups i.e. number of items along y axis.
    # These values are to be used for drawing the stair stepped
    # sorted P10-P90 area:

    coordinates = {"low": list(p10), "high": list(p90)}

    return coordinates



from typing import TYPE_CHECKING
import pathlib

import xtgeo
from webviz_config.webviz_store import webvizstore

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from ..property_statistics import PropertyStatistics


def surface_from_zone_prop(
    parent: "PropertyStatistics", zone: str, prop: str, ensemble: str, stype: str
) -> xtgeo.RegularSurface:

    if not isinstance(parent.surface_folders, dict):
        raise TypeError("parent.surface_folders must be of type dict")

    path = get_surface_path(
        ens_path=parent.surface_folders[ensemble],
        statistic=stype,
        zone=zone,
        prop=prop,
    )
    try:
        return xtgeo.surface_from_file(path.resolve())
    except OSError:
        surf = xtgeo.RegularSurface()
        surf.values = 0
        return surf


def surface_store(parent: "PropertyStatistics") -> list:
    """Function to copy all relevant statistical surfaces
    to store"""

    if not isinstance(parent.surface_folders, dict):
        raise TypeError("parent.surface_folders must be of type dict")

    zones = parent.pmodel.dataframe["ZONE"].unique()
    properties = parent.pmodel.dataframe["PROPERTY"].unique()
    ensembles = parent.pmodel.ensembles
    statistics = ["mean", "stddev", "min", "max", "p10", "p90"]
    store_funcs = []
    for ensemble in ensembles:
        for zone in zones:
            for prop in properties:
                for statistic in statistics:
                    store_funcs.append(
                        (
                            get_surface_path,
                            [
                                {
                                    "ens_path": parent.surface_folders[ensemble],
                                    "statistic": statistic,
                                    "zone": parent.surface_renaming.get(zone, zone),
                                    "prop": parent.surface_renaming.get(prop, prop),
                                }
                            ],
                        )
                    )
    return store_funcs


@webvizstore
def get_surface_path(
    ens_path: pathlib.Path, statistic: str, zone: str, prop: str
) -> pathlib.Path:
    return ens_path / statistic / f"{zone}--{prop}.gri"



from typing import Union, Optional, List, Tuple, Callable, Dict
import fnmatch
import pathlib
import re

import pandas as pd
from fmu.ensemble import ScratchEnsemble
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore


class EnsembleModel:
    """Class to load data from a scratchensemble using fmu.ensemble"""

    def __init__(
        self,
        ensemble_name: str,
        ensemble_path: Union[str, pathlib.Path],
        filter_file: Union[str, None] = "OK",
    ) -> None:
        self.ensemble_name = ensemble_name
        self.ensemble_path = str(ensemble_path)
        self.filter_file = filter_file
        self._webviz_store: List = []

    def __repr__(self) -> str:
        return f"EnsembleModel: {self.ensemble_name, self.ensemble_path}"

    @property
    def ens_folder(self) -> dict:
        """Get root folder for ensemble"""
        return {self.ensemble_name: self.ensemble_path.split("realization")[0]}

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def load_ensemble(self) -> ScratchEnsemble:
        ensemble = (
            ScratchEnsemble(self.ensemble_name, self.ensemble_path)
            if self.filter_file is None
            else ScratchEnsemble(self.ensemble_name, self.ensemble_path).filter(
                self.filter_file
            )
        )
        if ensemble.realizations == {}:
            raise ValueError(
                f"No realizations found for ensemble {self.ensemble_name}, "
                f"located at '{self.ensemble_path}'. "
                "Aborting..."
            )
        return ensemble

    def load_parameters(self) -> pd.DataFrame:
        self._webviz_store.append(
            (
                self._load_parameters,
                [{"self": self}],
            )
        )
        return self._load_parameters()

    def load_smry(
        self,
        time_index: Optional[Union[list, str]] = None,
        column_keys: Optional[list] = None,
    ) -> pd.DataFrame:
        # Limit saved columns if time_index = 'raw' or None, as the data density might
        # be very high, and increases risk of `MemoryError`
        if time_index == "raw" or time_index is None:
            self._webviz_store.append(
                (
                    self._load_smry,
                    [
                        {
                            "self": self,
                            "time_index": time_index,
                            "column_keys": column_keys,
                        }
                    ],
                )
            )
            return self._load_smry(time_index=time_index, column_keys=column_keys)

        # Otherwise store all columns to reduce risk of duplicates
        self._webviz_store.append(
            (
                self._load_smry,
                [{"self": self, "time_index": time_index, "column_keys": None}],
            )
        )

        if column_keys is None:
            return self._load_smry(time_index=time_index)
        df = self._load_smry(time_index=time_index)
        return df[
            df.columns[_match_column_keys(df_index=df.columns, column_keys=column_keys)]
        ]

    def load_smry_meta(
        self,
        column_keys: Optional[list] = None,
    ) -> pd.DataFrame:
        """Finds metadata for the summary vectors in the ensemble."""
        self._webviz_store.append(
            (self._load_smry_meta, [{"self": self, "column_keys": None}])
        )
        if column_keys is None:
            return self._load_smry_meta()
        df = self._load_smry_meta()
        return df[_match_column_keys(df_index=df.index, column_keys=column_keys)]

    def load_csv(self, csv_file: pathlib.Path) -> pd.DataFrame:
        self._webviz_store.append(
            (
                self._load_csv,
                [{"self": self, "csv_file": csv_file}],
            )
        )
        return self._load_csv(csv_file=csv_file)

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    @webvizstore
    def _load_parameters(self) -> pd.DataFrame:
        return self.load_ensemble().parameters

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    @webvizstore
    def _load_smry(
        self,
        time_index: Optional[Union[list, str]] = None,
        column_keys: Optional[list] = None,
    ) -> pd.DataFrame:
        return self.load_ensemble().get_smry(
            time_index=time_index, column_keys=column_keys
        )

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    @webvizstore
    def _load_smry_meta(
        self,
        column_keys: Optional[list] = None,
    ) -> pd.DataFrame:
        """Finds metadata for the summary vectors in the ensemble."""
        return pd.DataFrame(
            self.load_ensemble().get_smry_meta(column_keys=column_keys)
        ).T

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    @webvizstore
    def _load_csv(self, csv_file: pathlib.Path) -> pd.DataFrame:
        return self.load_ensemble().load_csv(str(csv_file))

    @property
    def webviz_store(self) -> List[Tuple[Callable, List[Dict]]]:
        return self._webviz_store


def _match_column_keys(
    df_index: pd.core.indexes.base.Index, column_keys: List[str]
) -> pd.core.indexes.base.Index:
    """Matches patterns in column_keys with the columns in df_columns, and adds 'DATE' and
    'REAL' to the requested column patterns.
    """
    all_columns_keys = ["DATE", "REAL"]
    all_columns_keys.extend(column_keys)
    regex = re.compile(
        "|".join([fnmatch.translate(column_key) for column_key in all_columns_keys])
    )
    return df_index.map(lambda column: bool(regex.fullmatch(column)))



from pathlib import Path
from typing import Optional, Union, Any

import pandas as pd
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore
from .fmu_input import load_ensemble_set

try:
    import ecl2df
except ImportError:
    pass

try:
    from pyscal import PyscalFactory
except ImportError:
    pass


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_satfunc(
    ensemble_paths: dict,
    ensemble_set_name: str = "EnsembleSet",
) -> pd.DataFrame:
    def ecl2df_satfunc(kwargs: Any) -> pd.DataFrame:
        return ecl2df.satfunc.df(kwargs["realization"].get_eclfiles())

    return load_ensemble_set(ensemble_paths, ensemble_set_name).apply(ecl2df_satfunc)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_scal_recommendation(
    scalfile: Path, sheet_name: Optional[Union[str, int, list]] = None
) -> pd.DataFrame:
    return PyscalFactory.create_scal_recommendation_list(
        PyscalFactory.load_relperm_df(str(scalfile), sheet_name)
    ).df()



from typing import Any
import pathlib

import pytest
import dash
from webviz_config.common_cache import CACHE
from _pytest.config.argparsing import Parser
from _pytest.fixtures import SubRequest


def pytest_addoption(parser: Parser) -> None:
    parser.addoption(
        "--testdata-folder",
        type=pathlib.Path,
        default=pathlib.Path("webviz-subsurface-testdata"),
        help="Path to webviz-subsurface-testdata folder",
    )


@pytest.fixture
def testdata_folder(request: SubRequest) -> Any:
    return request.config.getoption("--testdata-folder")


@pytest.fixture()
def app() -> dash.Dash:
    dash_app = dash.Dash(__name__)
    dash_app.css.config.serve_locally = True
    dash_app.scripts.config.serve_locally = True
    dash_app.config.suppress_callback_exceptions = True
    CACHE.init_app(dash_app.server)
    yield dash_app



from typing import Optional
import io
import base64

from PIL import Image
import numpy as np
import xtgeo

from webviz_subsurface._datainput.image_processing import array_to_png


class SurfaceLeafletModel:
    """Class to make a wcc.LeafletMap layer from a Xtgeo RegularSurface"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        surface: xtgeo.RegularSurface,
        name: Optional[str] = None,
        clip_min: Optional[float] = None,
        clip_max: Optional[float] = None,
        unit: str = " ",
        shader_type: str = "soft-hillshading",
        shadows: bool = False,
        colors: Optional[list] = None,
        updatemode: str = "update",
    ):
        self.name = name if name is not None else surface.name
        self.surface = surface
        self.shadows = True if shader_type == "hillshading_shadows" else shadows
        self.shader_type = (
            "hillshading" if shader_type == "hillshading_shadows" else shader_type
        )
        self.updatemode = updatemode
        self.bounds = [[surface.xmin, surface.ymin], [surface.xmax, surface.ymax]]
        self.zvalues = self.filled_z(clip_min=clip_min, clip_max=clip_max)
        self.unit = unit
        self.colors = self.set_colors(colors)

    @property
    def img_url(self) -> str:
        return array_to_png(self.zvalues.copy())

    @property
    def img_scale(self) -> float:
        img = Image.open(
            io.BytesIO(base64.b64decode(array_to_png(self.zvalues.copy())[22:]))
        )
        width, height = img.size
        if width * height >= 300 * 300:
            return 1.0
        ratio = (1000 ** 2) / (width * height)
        return np.sqrt(ratio).round(2)

    @property
    def min_val(self) -> float:
        return round(np.nanmin(self.zvalues), 4)

    @property
    def max_val(self) -> float:
        return round(np.nanmax(self.zvalues), 4)

    def filled_z(
        self,
        unrotate: bool = True,
        flip: bool = True,
        clip_min: float = None,
        clip_max: float = None,
    ) -> np.ndarray:
        surface = self.surface.copy()
        if clip_min or clip_max:
            np.ma.clip(surface.values, clip_min, clip_max, out=surface.values)
        if unrotate:
            surface.unrotate()

        x, y, z = surface.get_xyz_values()
        if flip:
            x = np.flip(x.transpose(), axis=0)
            y = np.flip(y.transpose(), axis=0)
            z = np.flip(z.transpose(), axis=0)
        return z.filled(np.nan)

    @staticmethod
    def set_colors(colors: list = None) -> list:
        return (
            [
                "#fde725",
                "#b5de2b",
                "#6ece58",
                "#35b779",
                "#1f9e89",
                "#26828e",
                "#31688e",
                "#3e4989",
                "#482878",
                "#440154",
            ]
            if colors is None
            else colors
        )

    @property
    def layer(self) -> dict:
        return {
            "name": self.name,
            "checked": True,
            "id": self.name,
            "action": self.updatemode,
            "baseLayer": True,
            "data": [
                {
                    "type": "image",
                    "url": self.img_url,
                    "colorScale": {
                        "colors": self.colors,
                        "prefixZeroAlpha": False,
                        "scaleType": "linear",
                    },
                    "shader": {
                        "type": self.shader_type,
                        "shadows": self.shadows,
                        "shadowIterations": 128,
                        "elevationScale": -1.0,
                        "pixelScale": 11000,
                        "setBlackToAlpha": True,
                    },
                    "bounds": self.bounds,
                    "minvalue": self.min_val,
                    "maxvalue": self.max_val,
                    "unit": self.unit,
                    "imageScale": self.img_scale,
                }
            ],
        }



from typing import List, Optional, Dict, Union, Tuple, Callable

import pandas as pd
import numpy as np
import dash
import dash_html_components as html
from dash.dependencies import Input, Output
import dash_core_components as dcc
import webviz_core_components as wcc
from webviz_config.common_cache import CACHE
from webviz_config import WebvizPluginABC

from webviz_subsurface._models import EnsembleSetModel
from .._utils.unique_theming import unique_colors
from .._utils.colors import hex_to_rgba


class BhpQc(WebvizPluginABC):
    """QC simulated bottom hole pressures (BHP) from reservoir simulations.

    Can be used to check if your simulated BHPs are in a realistic range.
    E.g. check if your simulated bottom hole pressures are very low in producers,
    or very high injectors.
    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    ---
    Data is read directly from the UNSMRY files with the raw frequency (not resampled).
    Resampling and csvs are not supported to avoid potential of interpolation, which
    might cover extreme BHP values.

    !> The `UNSMRY` files are auto-detected by `fmu-ensemble` in the `eclipse/model` folder of the \
    individual realizations. You should therefore not have more than one `UNSMRY` file in this \
    folder, to avoid risk of not extracting the right data.
    """

    def __init__(
        self,
        app: dash.Dash,
        ensembles: list,
        wells: Optional[List[str]] = None,
    ):
        super().__init__()
        if wells is None:
            self.column_keys = ["WBHP:*"]
        else:
            self.column_keys = [f"WBHP:{well}" for well in wells]

        self.emodel = EnsembleSetModel(
            ensemble_paths={
                ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][ens]
                for ens in ensembles
            }
        )
        self.smry = self.emodel.load_smry(
            time_index="raw", column_keys=self.column_keys
        )
        self.theme = app.webviz_settings["theme"]
        self.set_callbacks(app)

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": (
                    "Dashboard for BHP QC:"
                    "Check that simulated bottom hole pressures are realistic."
                ),
            },
            {"id": self.uuid("ensemble"), "content": "Select ensemble to QC."},
            {
                "id": self.uuid("sort_by"),
                "content": "Sort wells left to right according to this value.",
            },
            {
                "id": self.uuid("n_wells"),
                "content": (
                    "Show max selected number of top ranked wells after sorting and filtering."
                ),
            },
            {"id": self.uuid("wells"), "content": "Filter wells."},
        ]

    @property
    def ensembles(self) -> List[str]:
        return list(self.smry["ENSEMBLE"].unique())

    @property
    def wells(self) -> List[set]:
        return sorted(
            list(set(col[5:] for col in self.smry.columns if col.startswith("WBHP:")))
        )

    @property
    def ens_colors(self) -> dict:
        return unique_colors(self.ensembles, self.theme)

    @property
    def label_map(self) -> Dict[str, str]:
        return {
            "Mean": "mean",
            "Count (data points)": "count",
            "Stddev": "std",
            "Minimum": "min",
            "Maximum": "max",
            "P10 (high)": "high_p10",
            "P50": "p50",
            "P90 (low)": "low_p90",
        }

    @property
    def layout(self) -> wcc.FlexBox:
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                html.Div(
                    style={"flex": 1},
                    children=[
                        html.Label(
                            children=[
                                html.Span(
                                    "Ensemble:",
                                    style={"font-weight": "bold"},
                                ),
                                dcc.Dropdown(
                                    id=self.uuid("ensemble"),
                                    options=[
                                        {"label": i, "value": i} for i in self.ensembles
                                    ],
                                    value=self.ensembles[0],
                                    clearable=False,
                                    multi=False,
                                ),
                            ],
                        ),
                        html.Label(
                            children=[
                                html.Span("Plot type:", style={"font-weight": "bold"}),
                                dcc.Dropdown(
                                    id=self.uuid("plot_type"),
                                    options=[
                                        {"label": i, "value": i}
                                        for i in [
                                            "Fan chart",
                                            "Bar chart",
                                            "Line chart",
                                        ]
                                    ],
                                    clearable=False,
                                    value="Fan chart",
                                ),
                            ],
                        ),
                        html.Label(
                            id=self.uuid("select_stat"),
                            style={"display": "none"},
                            children=[
                                html.Span(
                                    "Select statistics:", style={"font-weight": "bold"}
                                ),
                                wcc.Select(
                                    id=self.uuid("stat_bars"),
                                    options=[
                                        {"label": key, "value": value}
                                        for key, value in self.label_map.items()
                                    ],
                                    size=8,
                                    value=["count", "low_p90", "p50"],
                                ),
                            ],
                        ),
                        html.Label(
                            children=[
                                html.Span("Sort by:", style={"font-weight": "bold"}),
                                dcc.Dropdown(
                                    id=self.uuid("sort_by"),
                                    options=[
                                        {"label": key, "value": value}
                                        for key, value in self.label_map.items()
                                    ],
                                    clearable=False,
                                    value="low_p90",
                                ),
                            ],
                        ),
                        dcc.RadioItems(
                            id=self.uuid("ascending"),
                            options=[
                                {"label": "Ascending", "value": True},
                                {"label": "Descending", "value": False},
                            ],
                            value=True,
                            labelStyle={"display": "inline-block"},
                        ),
                        html.Label(
                            children=[
                                html.Span(
                                    "Max number of wells in plot:",
                                    style={"font-weight": "bold"},
                                ),
                                dcc.Slider(
                                    id=self.uuid("n_wells"),
                                    min=1,
                                    max=len(self.wells),
                                    value=min(10, len(self.wells)),
                                    marks={1: 1, len(self.wells): len(self.wells)},
                                ),
                            ]
                        ),
                        html.Label(
                            children=[
                                html.Span("Wells:", style={"font-weight": "bold"}),
                                wcc.Select(
                                    id=self.uuid("wells"),
                                    options=[
                                        {"label": i, "value": i} for i in self.wells
                                    ],
                                    size=min([len(self.wells), 20]),
                                    value=self.wells,
                                ),
                            ],
                        ),
                    ],
                ),
                html.Div(
                    style={"flex": 3},
                    children=[
                        html.Div(
                            # style={"height": "300px"},
                            children=wcc.Graph(id=self.uuid("graph")),
                        ),
                    ],
                ),
            ],
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            Output(self.uuid("graph"), "figure"),
            Input(self.uuid("ensemble"), "value"),
            Input(self.uuid("plot_type"), "value"),
            Input(self.uuid("n_wells"), "value"),
            Input(self.uuid("wells"), "value"),
            Input(self.uuid("sort_by"), "value"),
            Input(self.uuid("stat_bars"), "value"),
            Input(self.uuid("ascending"), "value"),
        )
        def _update_graph(
            ensemble: str,
            plot_type: str,
            n_wells: int,
            wells: Union[str, List[str]],
            sort_by: str,
            stat_bars: Union[str, List[str]],
            ascending: bool,
        ) -> dict:
            wells = wells if isinstance(wells, list) else [wells]
            stat_bars = stat_bars if isinstance(stat_bars, list) else [stat_bars]
            df = filter_df(df=self.smry, ensemble=ensemble, wells=wells)
            stat_df = (
                calc_statistics(df)
                .sort_values(sort_by, ascending=ascending)
                .iloc[0:n_wells, :]
            )
            traces = []
            if plot_type == "Fan chart":
                traces.extend(
                    add_fanchart_traces(
                        ens_stat_df=stat_df,
                        color=self.ens_colors[ensemble],
                        legend_group=ensemble,
                    )
                )
            elif plot_type in ["Bar chart", "Line chart"]:
                for stat in stat_bars:
                    yaxis = "y2" if stat == "count" else "y"

                    if plot_type == "Bar chart":

                        traces.append(
                            {
                                "x": [vec[5:] for vec in stat_df.index],  # strip WBHP:
                                "y": stat_df[stat],
                                "name": [
                                    key
                                    for key, value in self.label_map.items()
                                    if value == stat
                                ][0],
                                "yaxis": yaxis,
                                "type": "bar",
                                "offsetgroup": stat,
                                "showlegend": True,
                            }
                        )
                    elif plot_type == "Line chart":
                        traces.append(
                            {
                                "x": [vec[5:] for vec in stat_df.index],  # strip WBHP:
                                "y": stat_df[stat],
                                "name": [
                                    key
                                    for key, value in self.label_map.items()
                                    if value == stat
                                ][0],
                                "yaxis": yaxis,
                                "type": "line",
                                "offsetgroup": stat,
                                "showlegend": True,
                            }
                        )
                    else:
                        raise ValueError("Invalid plot type.")

            layout = self.theme.create_themed_layout(
                {
                    "yaxis": {
                        "side": "left",
                        "title": "Bottom hole pressure",
                        "showgrid": False,
                    },
                    "yaxis2": {
                        "side": "right",
                        "overlaying": "y",
                        "title": "Count (data points)",
                        "showgrid": False,
                    },
                    "xaxis": {"showgrid": False},
                    "barmode": "group",
                    "legend": {"x": 1.05},
                }
            )
            return {"data": traces, "layout": layout}

        @app.callback(
            Output(self.uuid("select_stat"), "style"),
            Input(self.uuid("plot_type"), "value"),
        )
        def _update_stat_selector(plot_type: str) -> dict:
            return (
                {"display": "none"}
                if plot_type == "Fan chart"
                else {"display": "block"}
            )

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return self.emodel.webvizstore


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_df(df: pd.DataFrame, ensemble: str, wells: List[str]) -> pd.DataFrame:
    """Filter dataframe for current ensembles and wells.
    Replacing zeroes (well not open) with np.NaN to not be accounted for
    in statistics.
    """
    columns = ["ENSEMBLE"] + [f"WBHP:{well}" for well in wells]
    return df.loc[df["ENSEMBLE"] == ensemble][columns].replace(0, np.NaN)


def calc_statistics(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate statistics for given vectors over the ensembles
    refaxis is used if another column than DATE should be used to groupby.
    """
    # Invert p10 and p90 due to oil industry convention.
    def p10(x: np.array) -> float:
        return np.nanpercentile(x, q=90)

    def p50(x: np.array) -> float:
        return np.nanpercentile(x, q=50)

    def p90(x: np.array) -> float:
        return np.nanpercentile(x, q=10)

    # Calculate statistics, ignoring NaNs.
    stat_df = (
        df.groupby("ENSEMBLE")
        .agg([np.nanmean, "count", np.nanstd, np.nanmin, np.nanmax, p10, p50, p90])
        .reset_index(drop=True, level="ENSEMBLE")
    )
    # Rename nanmin, nanmax and nanmean to min, max and mean.
    col_stat_label_map = {
        "nanmin": "min",
        "nanmax": "max",
        "nanmean": "mean",
        "nanstd": "std",
        "p10": "high_p10",
        "p90": "low_p90",
    }
    stat_df.rename(columns=col_stat_label_map, level=1, inplace=True)
    stat_df = stat_df.transpose().unstack()
    stat_df.columns = stat_df.columns.get_level_values(1)  # Remove 0 index column
    return stat_df


def add_fanchart_traces(
    ens_stat_df: pd.DataFrame,
    color: str,
    legend_group: str,
) -> List[dict]:
    """Renders a fanchart for an ensemble vector"""

    fill_color = hex_to_rgba(color, 0.3)
    line_color = hex_to_rgba(color, 1)
    x = [vec[5:] for vec in ens_stat_df.index]

    return [
        {
            "name": legend_group,
            "hovertext": "Maximum",
            "x": x,
            "y": ens_stat_df["max"],
            "mode": "lines",
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
        {
            "name": legend_group,
            "hovertext": "P90 (low)",
            "x": x,
            "y": ens_stat_df["low_p90"],
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
        {
            "name": legend_group,
            "hovertext": "P10 (high)",
            "x": x,
            "y": ens_stat_df["high_p10"],
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
        {
            "name": legend_group,
            "hovertext": "Mean",
            "x": x,
            "y": ens_stat_df["mean"],
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"color": line_color},
            "legendgroup": legend_group,
            "showlegend": True,
        },
        {
            "name": legend_group,
            "hovertext": "Minimum",
            "x": x,
            "y": ens_stat_df["min"],
            "mode": "lines",
            "fill": "tonexty",
            "fillcolor": fill_color,
            "line": {"width": 0, "color": line_color},
            "legendgroup": legend_group,
            "showlegend": False,
        },
    ]



from typing import Union, Optional

import dash_html_components as html
import dash_core_components as dcc
import webviz_core_components as wcc

from ..figures.color_figure import color_figure


def ensemble_selector(
    parent,
    tab: str,
    id_string: str,
    multi: bool = False,
    value: str = None,
    heading: str = None,
) -> html.Div:
    children = []
    if heading is not None:
        children.append(html.Span(heading, style={"font-weight": "bold"}))
    children.append(
        dcc.Dropdown(
            id={"id": parent.uuid(id_string), "tab": tab},
            options=[{"label": ens, "value": ens} for ens in parent.pmodel.ensembles],
            multi=multi,
            value=value if value is not None else parent.pmodel.ensembles[0],
            clearable=False,
            persistence=True,
            persistence_type="session",
        )
    )
    return html.Div(style={"width": "90%"}, children=children)


def vector_selector(parent) -> html.Div:
    return html.Div(
        style={"width": "90%", "margin-top": "15px"},
        children=[
            html.Span("Vector type:", style={"font-weight": "bold"}),
            html.Div(
                id=parent.uuid("vtype-container"),
                children=[
                    dcc.RadioItems(
                        id={"id": parent.uuid("vtype-select"), "state": "initial"},
                        options=[
                            {"label": i, "value": i}
                            for i in parent.vmodel.vector_groups
                        ],
                        value="Field",
                        labelStyle={"display": "inline-block", "margin-right": "10px"},
                    )
                ],
            ),
            html.Div(
                style={"margin-top": "5px"},
                children=[
                    html.Span("Vector:", style={"font-weight": "bold"}),
                    html.Div(
                        id=parent.uuid("vshort-container"),
                        children=[
                            dcc.Dropdown(
                                id=parent.uuid("vshort-select"),
                                options=[
                                    {"label": i, "value": i}
                                    for i in parent.vmodel.vector_groups["Field"][
                                        "shortnames"
                                    ]
                                ],
                                value=parent.vmodel.vector_groups["Field"][
                                    "shortnames"
                                ][0],
                                placeholder="Select a vector...",
                                clearable=False,
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                ],
            ),
            html.Div(id=parent.uuid("vitems-container"), children=[]),
            html.Div(id=parent.uuid("vector-select"), style={"display": "none"}),
            html.Div(id=parent.uuid("clickdata-store"), style={"display": "none"}),
        ],
    )


def parameter_selector(parent, tab: str) -> html.Div:
    return html.Div(
        style={"width": "90%", "margin-top": "15px"},
        children=[
            html.Span("Parameter:", style={"font-weight": "bold"}),
            dcc.Dropdown(
                id={"id": parent.uuid("parameter-select"), "tab": tab},
                options=[{"label": i, "value": i} for i in parent.pmodel.parameters],
                placeholder="Select a parameter...",
                clearable=False,
                persistence=True,
                persistence_type="session",
            ),
        ],
    )


def sortby_selector(
    parent,
    value: str = None,
) -> html.Div:
    return html.Div(
        style={"margin-top": "10px", "width": "90%"},
        children=[
            html.Span(
                "Sort parameters by:",
                style={"font-weight": "bold"},
            ),
            dcc.RadioItems(
                id=parent.uuid("delta-sort"),
                options=[
                    {"label": "Name", "value": "Name"},
                    {
                        "label": "StdDev",
                        "value": "Stddev",
                    },
                    {
                        "label": "Average",
                        "value": "Avg",
                    },
                ],
                value=value,
                labelStyle={
                    "display": "inline-block",
                    "margin-right": "5px",
                },
                persistence=True,
                persistence_type="session",
            ),
        ],
    )


def plot_options(parent, tab: str) -> html.Div:
    return html.Div(
        style={"width": "90%", "margin-top": "15px"},
        children=[
            dcc.Checklist(
                id={"id": parent.uuid("checkbox-options"), "tab": tab},
                options=[
                    {"label": "Dateline visible", "value": "DateLine"},
                    {"label": "Auto compute correlations", "value": "AutoCompute"},
                ],
                value=["DateLine", "AutoCompute"],
            ),
            html.Div(
                id={"id": parent.uuid("plot-options"), "tab": tab},
                style={"display": "none"},
            ),
        ],
    )


def date_selector(parent) -> html.Div:
    dates = parent.vmodel.dates
    return html.Div(
        style={"width": "90%", "margin-top": "15px"},
        children=[
            html.Div(
                style={"display": "inline-flex"},
                children=[
                    html.Span("Date:", style={"font-weight": "bold"}),
                    html.Span(
                        "date",
                        id=parent.uuid("date-selected"),
                        style={"margin-left": "10px"},
                    ),
                ],
            ),
            dcc.Slider(
                id=parent.uuid("date-slider"),
                value=len(dates) - 1,
                min=0,
                max=len(dates) - 1,
                included=False,
                marks={
                    idx: {
                        "label": dates[idx],
                        "style": {
                            "white-space": "nowrap",
                        },
                    }
                    for idx in [0, len(dates) - 1]
                },
            ),
        ],
    )


def filter_parameter(
    parent,
    tab: str,
    multi: bool = True,
    value: Union[str, float] = None,
) -> html.Div:
    return html.Div(
        style={"margin-top": "10px", "width": "90%"},
        children=[
            html.Span("Parameters:", style={"font-weight": "bold"}),
            html.Div(
                id=parent.uuid("filter-parameter-container"),
                children=[
                    wcc.Select(
                        id={
                            "id": parent.uuid("filter-parameter"),
                            "tab": tab,
                        },
                        options=[
                            {"label": i, "value": i} for i in parent.pmodel.parameters
                        ],
                        value=value,
                        multi=multi,
                        size=min(40, len(parent.pmodel.parameters)),
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
        ],
    )


def make_filter(
    parent,
    tab: str,
    vtype: str,
    column_values: list,
    multi: bool = True,
    value: Union[str, float] = None,
    open_details: bool = False,
) -> html.Div:
    return html.Div(
        children=html.Details(
            open=open_details,
            children=[
                html.Summary(vtype),
                wcc.Select(
                    id={
                        "id": parent.uuid("vitem-filter"),
                        "tab": tab,
                        "vtype": vtype,
                    },
                    options=[{"label": i, "value": i} for i in column_values],
                    value=[value] if value is not None else column_values,
                    multi=multi,
                    size=min(15, len(column_values)),
                    persistence=True,
                    persistence_type="session",
                ),
            ],
        ),
    )


def filter_vector_selector(
    parent,
    tab: str,
    multi: bool = True,
    value: Union[str, float] = None,
    open_details: bool = False,
) -> html.Div:
    return html.Div(
        style={"width": "90%"},
        children=[
            html.Span("Vector type:", style={"font-weight": "bold"}),
            dcc.Dropdown(
                id={
                    "id": parent.uuid("vtype-filter"),
                    "tab": tab,
                },
                options=[{"label": i, "value": i} for i in parent.vmodel.vector_groups],
                value=list(parent.vmodel.vector_groups),
                clearable=False,
                style={"background-color": "white"},
                multi=True,
                persistence=True,
                persistence_type="session",
            ),
            html.Div(
                children=[
                    make_filter(
                        parent=parent,
                        tab=tab,
                        vtype=f"{vtype}s",
                        column_values=vlist["items"],
                        multi=multi,
                        value=value,
                        open_details=open_details,
                    )
                    for vtype, vlist in parent.vmodel.vector_groups.items()
                    if vlist["items"]
                ],
            ),
            html.Div(
                id={"id": parent.uuid("filter-select"), "tab": tab},
                style={"display": "none"},
            ),
        ],
    )


def color_selector(
    parent,
    tab: str,
    colors: Optional[list] = None,
    bargap: Optional[float] = None,
    height: Optional[float] = None,
):
    return html.Div(
        style={"width": "90%", "margin-top": "5px"},
        children=[
            html.Span("Colors:", style={"font-weight": "bold"}),
            wcc.Graph(
                id={"id": parent.uuid("color-selector"), "tab": tab},
                config={"displayModeBar": False},
                figure=color_figure(
                    colors=colors,
                    bargap=bargap,
                    height=height,
                ),
            ),
        ],
    )


def color_opacity_selector(parent, tab: str, value: float):
    return html.Div(
        style={"width": "90%", "margin-top": "5px"},
        children=[
            "Opacity:",
            dcc.Input(
                id={"id": parent.uuid("opacity-selector"), "tab": tab},
                type="number",
                min=0,
                max=1,
                step=0.1,
                value=value,
                style={"margin-left": "10px"},
            ),
        ],
    )


def html_details(
    summary: str,
    children: list,
    open_details: bool = False,
) -> html.Div:
    return html.Div(
        html.Details(
            style={"margin-bottom": "25px"},
            open=open_details,
            children=[
                html.Summary(
                    children=summary,
                    style={
                        "color": "#ff1243",
                        "border-bottom-style": "solid",
                        "border-width": "thin",
                        "border-color": "#ff1243",
                        "font-weight": "bold",
                        "font-size": "20px",
                        "margin-bottom": "15px",
                    },
                )
            ]
            + children,
        )
    )



from typing import List, Dict, Any

import pandas as pd


class MapFigure:
    def __init__(self, ertdf: pd.DataFrame, ensemble: str) -> None:

        self.ertdf = (
            ertdf.loc[ertdf["ENSEMBLE"] == ensemble]
            .groupby(["WELL", "DATE", "ENSEMBLE"])
            .aggregate("mean")
            .reset_index()
        )
        self.traces: List[Dict[str, Any]] = []

    def add_misfit_plot(self, sizeby: str, colorby: str, dates: List[float]) -> None:
        df = self.ertdf.loc[
            (self.ertdf["DATE_IDX"] >= dates[0]) & (self.ertdf["DATE_IDX"] <= dates[1])
        ]

        self.traces.append(
            {
                "x": df["EAST"],
                "y": df["NORTH"],
                "text": df["WELL"],
                "customdata": df["WELL"],
                "mode": "markers",
                "hovertext": [
                    f"Well: {well}"
                    f"<br>Mean simulated pressure: {pressure:.2f}"
                    f"<br>Mean misfit: {misfit:.2f}"
                    f"<br>Stddev pressure: {stddev:.2f}"
                    for well, stddev, misfit, pressure in zip(
                        df["WELL"], df["STDDEV"], df["DIFF"], df["SIMULATED"]
                    )
                ],
                "hoverinfo": "text",
                # "name": date,
                "showlegend": False,
                "marker": {
                    "size": df[sizeby],
                    "sizeref": 2.0 * self.ertdf[sizeby].quantile(0.9) / (40.0 ** 2),
                    "sizemode": "area",
                    "sizemin": 6,
                    "color": df[colorby],
                    "cmin": self.ertdf[colorby].min(),
                    "cmax": self.ertdf[colorby].quantile(0.9),
                    "colorscale": [[0, "#2584DE"], [1, "#E50000"]],
                    "showscale": True,
                    "colorbar": {"x": -0.1},
                },
            }
        )

    def add_fault_lines(self, df: pd.DataFrame) -> None:
        for _fault, faultdf in df.groupby("POLY_ID"):
            self.traces.append(
                {
                    "x": faultdf["X_UTME"],
                    "y": faultdf["Y_UTMN"],
                    "mode": "lines",
                    "type": "scatter",
                    "hoverinfo": "none",
                    "showlegend": False,
                    "line": {"color": "grey", "width": 1},
                }
            )

    @property
    def layout(self) -> Dict[str, Any]:
        """The plotly figure layout"""
        return {
            "hovermode": "closest",
            "legend": {"itemsizing": "constant", "orientation": "h"},
            "height": 800,
            "colorway": ["red", "blue"],
            "margin": {"t": 50, "l": 0, "r": 0},
            "xaxis": {"constrain": "domain", "showgrid": False},
            "yaxis": {"scaleanchor": "x", "showgrid": False},
        }



from typing import List, Dict, Tuple, Callable
from uuid import uuid4
from pathlib import Path

import pandas as pd
import dash
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
import dash_html_components as html
import dash_core_components as dcc
import webviz_subsurface_components as wsc
from webviz_config import WebvizPluginABC
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from .._datainput.fmu_input import load_parameters


class ParameterDistribution(WebvizPluginABC):
    """Visualizes parameter distributions for FMU ensembles.

Parameters are visualized either as histograms, showing parameter ranges
and distributions for each ensemble.

Input can be given either as an aggregated `csv` file with parameter information,
or as ensemble name(s) defined in `shared_settings`.

---

**Using aggregated data**
* **`csvfile`:** Aggregated `csv` file with `REAL`, `ENSEMBLE` and parameter columns. \
 (absolute path or relative to config file).

**Reading data from ensembles**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.

---
Parameter values are extracted automatically from the `parameters.txt` files in the individual
realizations if you have defined `ensembles`, using the `fmu-ensemble` library.

When using an aggregated `csvfile`, you need to have the columns `REAL`, `ENSEMBLE`
and the parameter columns.
"""

    def __init__(self, app: dash.Dash, csvfile: Path = None, ensembles: list = None):

        super().__init__()

        self.csvfile = csvfile if csvfile else None

        if csvfile and ensembles:
            raise ValueError(
                'Incorrect arguments. Either provide a "csvfile" or "ensembles".'
            )
        if csvfile:
            self.parameters = read_csv(csvfile)
        elif ensembles:
            self.ensembles = {
                ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.parameters = load_parameters(
                ensemble_paths=self.ensembles, ensemble_set_name="EnsembleSet"
            )
        else:
            raise ValueError(
                'Incorrect arguments. Either provide a "csvfile" or "ensembles".'
            )

        self.parameter_columns: List[str] = [
            col
            for col in list(self.parameters.columns)
            if col not in ["REAL", "ENSEMBLE"]
        ]
        self.uid = uuid4()
        self.plotly_theme = app.webviz_settings["theme"].plotly_theme
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.ids("layout"),
                "content": ("Dashboard displaying distribution of input parameters"),
            },
            {
                "id": self.ids("graph"),
                "content": (
                    "Visualization of currently selected parameter as histogram "
                    "series and distribution range per ensemble."
                ),
            },
            {
                "id": self.ids("parameter"),
                "content": (
                    "Select visualized parameter by selecting or searching the list."
                ),
            },
        ]

    @staticmethod
    def set_grid_layout(columns: str) -> Dict[str, str]:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    def make_buttons(self, prev_id: str, next_id: str) -> html.Div:
        return html.Div(
            style=self.set_grid_layout("1fr 1fr"),
            children=[
                html.Button(
                    id=prev_id,
                    style={
                        "fontSize": "2rem",
                        "paddingLeft": "5px",
                        "paddingRight": "5px",
                    },
                    children="⬅",
                ),
                html.Button(
                    id=next_id,
                    style={
                        "fontSize": "2rem",
                        "paddingLeft": "5px",
                        "paddingRight": "5px",
                    },
                    children="➡",
                ),
            ],
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            id=self.ids("layout"),
            children=[
                html.Span("Parameter distribution:", style={"font-weight": "bold"}),
                html.Div(
                    style=self.set_grid_layout("8fr 1fr 2fr"),
                    children=[
                        dcc.Dropdown(
                            id=self.ids("parameter"),
                            options=[
                                {"value": col, "label": col}
                                for col in self.parameter_columns
                            ],
                            value=self.parameter_columns[0],
                            clearable=False,
                            persistence=True,
                            persistence_type="session",
                        ),
                        self.make_buttons(self.ids("prev-btn"), self.ids("next-btn")),
                    ],
                ),
                wsc.PriorPosteriorDistribution(id=self.ids("graph")),
            ],
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            Output(self.ids("parameter"), "value"),
            [
                Input(self.ids("prev-btn"), "n_clicks"),
                Input(self.ids("next-btn"), "n_clicks"),
            ],
            [State(self.ids("parameter"), "value")],
        )
        def _set_parameter_from_btn(
            _prev_click: int, _next_click: int, column: str
        ) -> str:

            ctx = dash.callback_context.triggered
            if ctx is None:
                raise PreventUpdate
            callback = ctx[0]["prop_id"]
            if callback == f"{self.ids('prev-btn')}.n_clicks":
                column = prev_value(column, self.parameter_columns)
            elif callback == f"{self.ids('next-btn')}.n_clicks":
                column = next_value(column, self.parameter_columns)
            else:
                column = self.parameter_columns[0]
            return column

        @app.callback(
            Output(self.ids("graph"), "data"), [Input(self.ids("parameter"), "value")]
        )
        def _set_parameter(column: str) -> dict:
            param = self.parameters[[column, "REAL", "ENSEMBLE"]]

            ensembles = param["ENSEMBLE"].unique().tolist()

            iterations = []
            values = []
            labels = []

            for ensemble in ensembles:
                df = param[param["ENSEMBLE"] == ensemble]
                iterations.append(ensemble)
                values.append(df[column].tolist())
                labels.append([f"Realization {real}" for real in df["REAL"].tolist()])

            return {"iterations": iterations, "values": values, "labels": labels}

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (read_csv, [{"csv_file": self.csvfile}])
            if self.csvfile
            else (
                load_parameters,
                [
                    {
                        "ensemble_paths": self.ensembles,
                        "ensemble_set_name": "EnsembleSet",
                    }
                ],
            )
        ]


def prev_value(current_value: str, options: List[str]) -> str:
    try:
        index = options.index(current_value)
        return options[max(0, index - 1)]
    except ValueError:
        return current_value


def next_value(current_value: str, options: List[str]) -> str:
    try:
        index = options.index(current_value)
        return options[min(len(options) - 1, index + 1)]
    except ValueError:
        return current_value


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)



from typing import Dict, List, Any

import xtgeo
import numpy as np

from webviz_config.common_cache import CACHE


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def load_well(well_path: str) -> xtgeo.Well:
    return xtgeo.Well(well_path)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def make_well_layer(
    well: xtgeo.Well, name: str = "well", zmin: float = 0
) -> Dict[str, Any]:
    """Make LayeredMap well polyline"""
    well.dataframe = well.dataframe[well.dataframe["Z_TVDSS"] > zmin]
    positions = well.dataframe[["X_UTME", "Y_UTMN"]].values
    return {
        "name": name,
        "checked": True,
        "base_layer": False,
        "data": [
            {
                "type": "polyline",
                "color": "black",
                "positions": positions,
                "tooltip": name,
            }
        ],
    }


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def make_well_layers(
    wellfiles: List[str], zmin: float = 0, max_points: float = 100
) -> Dict[str, Any]:
    """Make layeredmap wells layer"""
    data = []
    for wellfile in wellfiles:
        try:
            well = load_well(wellfile)
        except ValueError:
            continue
        well.dataframe = well.dataframe[well.dataframe["Z_TVDSS"] > zmin]
        while len(well.dataframe.values) > max_points:
            well.downsample()
        positions = well.dataframe[["X_UTME", "Y_UTMN"]].values
        data.append(
            {
                "type": "polyline",
                "color": "black",
                "positions": positions,
                "tooltip": well.name,
            }
        )
    return {"name": "Wells", "checked": True, "base_layer": False, "data": data}


def get_well_layers(
    wells: Dict[str, xtgeo.Well],
    planned_wells: Dict[str, xtgeo.Well],
    surface_name: str,
    surface: xtgeo.RegularSurface,
    dropdown_file: str,
) -> List[Dict[str, Any]]:
    """Make circles around well in layered map view
    Args:
        wells: dictionary of  type {wellfile: xtgeo.Well(wellfile)}
        planned_wells: dictionary of type {wellfile: xtgeo.Well(wellfile)}
        surface_name: Name of surface
        surface: xtgeo surface
        dropdown_file: A well file from dropdown value
    Returns:
        List of well layers with data for well circles and trajectory
    """
    data: List[Dict[str, Any]] = []
    planned_data: List[Dict[str, Any]] = []
    dropdown_well = (
        wells[dropdown_file] if dropdown_file in wells else planned_wells[dropdown_file]
    )
    dropdown_well_df = dropdown_well.dataframe[dropdown_well.dataframe["Z_TVDSS"] > 0]
    positions = dropdown_well_df[["X_UTME", "Y_UTMN"]].values
    dropdown_data = [
        {
            "type": "polyline",
            "color": "rgba(128,0,0)",
            "positions": positions,
            "tooltip": dropdown_well.wellname + " trajectory",
        }
    ]
    for wellfile, well in wells.items():
        color = "rgba(128,0,0)" if wellfile == dropdown_file else "rgba(255,20,147)"
        append_well_to_data(data, well, wellfile, surface, color)
    for wellfile, well in planned_wells.items():
        color = "rgba(128,0,0)" if wellfile == dropdown_file else "rgba(224,224,224,1)"
        append_well_to_data(planned_data, well, wellfile, surface, color)
    return [
        {
            "name": "Wells",
            "checked": True,
            "baseLayer": False,
            "data": data,
            "id": surface_name + "-wells",
            "action": "add",
        },
        {
            "name": "Planned wells",
            "checked": True,
            "baselayer": False,
            "data": planned_data,
            "id": surface_name + "-planned-wells",
            "action": "add",
        },
        {
            "name": dropdown_well.wellname + "-trajectory",
            "checked": True,
            "baseLayer": False,
            "data": dropdown_data,
            "action": "add",
            "id": surface_name + "-well-trajectory",
        },
    ]


def append_well_to_data(
    data: List[Dict[str, Any]],
    well: xtgeo.Well,
    wellfile: str,
    surface: str,
    color: str,
) -> None:
    with np.errstate(invalid="ignore"):
        surface_picks = well.get_surface_picks(surface)
        # get_surface_picks raises warning when MD column is missing in well
    if surface_picks is not None:
        surface_picks_df = surface_picks.dataframe
        coordinates = surface_picks_df[["X_UTME", "Y_UTMN"]].values
        for coord in coordinates:
            data.append(
                {
                    "type": "circle",
                    "center": coord,
                    "color": color,
                    "radius": 100,
                    "tooltip": well.wellname,
                    "id": wellfile,
                }
            )



from typing import List, Union, Tuple, Callable
from pathlib import Path
import json
import io
import warnings

import numpy as np
import pandas as pd
import xtgeo
import dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_html_components as html
import dash_core_components as dcc
from webviz_subsurface_components import LeafletMap
import webviz_core_components as wcc
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import CACHE
from webviz_config import WebvizPluginABC

from webviz_subsurface._datainput.fmu_input import get_realizations, find_surfaces
from webviz_subsurface._datainput.surface import make_surface_layer, load_surface
from webviz_subsurface._datainput.well import make_well_layers
from webviz_subsurface._private_plugins.surface_selector import SurfaceSelector


class SurfaceViewerFMU(WebvizPluginABC):
    """Covisualize surfaces from an ensemble.

There are 3 separate map views. 2 views can be set independently, while
the 3rd view displays the resulting map by combining the other maps, e.g.
by taking the difference or summing the values.

There is flexibility in which combinations of surfaces that are displayed
and calculated, such that surfaces can be compared across ensembles and realizations.

Statistical calculations across the ensemble(s) are
done on the fly. If the ensemble(s) or surfaces have a large size, it is recommended
to run webviz in `portable` mode so that the statistical surfaces are pre-calculated,
and available for instant viewing.

---

* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`attributes`:** List of surface attributes to include, if not given
    all surface attributes will be included.
* **`attribute_settings`:** Dictionary with setting for each attribute.
    Available settings are:
    * `min`: Truncate colorscale (lower limit).
    * `max`: Truncate colorscale (upper limit).
    * `color`: List of hexadecimal colors.
    * `unit`: Text to display as unit in label.
* **`wellfolder`:** Folder with RMS wells.
* **`wellsuffix`:** File suffix for wells in well folder.
* **`map_height`:** Set the height in pixels for the map views.

---
The available maps are gathered from the `share/results/maps/` folder
for each realization. Subfolders are not supported.

The filenames need to follow a fairly strict convention, as the filenames are used as metadata:
`horizon_name--attribute--date` (`--date` is optional). The files should be on `irap binary`
format with the suffix `.gri`. The date is of the form `YYYYMMDD` or
`YYYYMMDD_YYYYMMDD`, the latter would be for a delta surface between two dates.
See [this folder]\
(https://github.com/equinor/webviz-subsurface-testdata/tree/master/reek_history_match/\
realization-0/iter-0/share/results/maps) \
for examples of file naming conventions.

The `attribute_settings` consists of optional settings for the individual attributes that are
extracted based on the filenames mentioned above. For attributes called `atr_a` and `atr_b`, the
configuration of `attribute_settings` could e.g. be:
```yaml
attribute_settings:
  atr_a:
    min: 4
    max: 10
    unit: m
  atr_b:
    color:
    - "#000004"
    - "#1b0c41"
    - "#4a0c6b"
    - "#781c6d"
```

"""

    def __init__(
        self,
        app: dash.Dash,
        ensembles: list,
        attributes: list = None,
        attribute_settings: dict = None,
        wellfolder: Path = None,
        wellsuffix: str = ".w",
        map_height: int = 600,
    ):

        super().__init__()
        self.ens_paths = {
            ens: app.webviz_settings["shared_settings"]["scratch_ensembles"][ens]
            for ens in ensembles
        }

        # Find surfaces
        self.surfacedf = find_surfaces(self.ens_paths)
        if attributes is not None:
            self.surfacedf = self.surfacedf[
                self.surfacedf["attribute"].isin(attributes)
            ]
            if self.surfacedf.empty:
                raise ValueError("No surfaces found with the given attributes")
        self.attribute_settings: dict = attribute_settings if attribute_settings else {}
        self.map_height = map_height
        self.surfaceconfig = surfacedf_to_dict(self.surfacedf)
        self.wellfolder = wellfolder
        self.wellsuffix = wellsuffix
        self.wellfiles: Union[List[str], None] = (
            json.load(find_files(wellfolder, wellsuffix))
            if wellfolder is not None
            else None
        )
        self.well_layer = (
            make_well_layers([get_path(wellfile) for wellfile in self.wellfiles])
            if self.wellfiles
            else None
        )
        # Extract realizations and sensitivity information
        self.ens_df = get_realizations(
            ensemble_paths=self.ens_paths, ensemble_set_name="EnsembleSet"
        )
        self.selector = SurfaceSelector(app, self.surfaceconfig, ensembles)
        self.selector2 = SurfaceSelector(app, self.surfaceconfig, ensembles)

        self.set_callbacks(app)

    @property
    def ensembles(self) -> List[str]:
        return list(self.ens_df["ENSEMBLE"].unique())

    def realizations(
        self, ensemble: str, sensname: str = None, senstype: str = None
    ) -> List[str]:
        df = self.ens_df.loc[self.ens_df["ENSEMBLE"] == ensemble].copy()
        if sensname and senstype:
            df = df.loc[(df["SENSNAME"] == sensname) & (df["SENSCASE"] == senstype)]
        return list(df["REAL"]) + ["Mean", "StdDev", "Min", "Max"]

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": (
                    "Dashboard to compare surfaces from a FMU ensemble. "
                    "The two left views can be set independently, while the right "
                    "view shows a calculated surface."
                ),
            },
            {
                "id": self.uuid("settings-view1"),
                "content": ("Settings for the first map view"),
            },
            {
                "id": self.uuid("settings-view2"),
                "content": ("Settings for the second map view"),
            },
            {
                "id": self.uuid("settings-view3"),
                "content": ("Settings for the calculated map view"),
            },
        ]

    @staticmethod
    def set_grid_layout(columns: str) -> dict:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    def ensemble_layout(
        self,
        ensemble_id: str,
        ens_prev_id: str,
        ens_next_id: str,
        real_id: str,
        real_prev_id: str,
        real_next_id: str,
    ) -> wcc.FlexBox:
        return wcc.FlexBox(
            children=[
                html.Div(
                    [
                        html.Label("Ensemble"),
                        html.Div(
                            style=self.set_grid_layout("12fr 1fr 1fr"),
                            children=[
                                dcc.Dropdown(
                                    options=[
                                        {"label": ens, "value": ens}
                                        for ens in self.ensembles
                                    ],
                                    value=self.ensembles[0],
                                    id=ensemble_id,
                                    clearable=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                                html.Button(
                                    style={
                                        "fontSize": "2rem",
                                        "paddingLeft": "5px",
                                        "paddingRight": "5px",
                                    },
                                    id=ens_prev_id,
                                    children="⬅",
                                ),
                                html.Button(
                                    style={
                                        "fontSize": "2rem",
                                        "paddingLeft": "5px",
                                        "paddingRight": "5px",
                                    },
                                    id=ens_next_id,
                                    children="➡",
                                ),
                            ],
                        ),
                    ]
                ),
                html.Div(
                    children=[
                        html.Label("Realization / Statistic"),
                        html.Div(
                            style=self.set_grid_layout("12fr 1fr 1fr"),
                            children=[
                                dcc.Dropdown(
                                    options=[
                                        {"label": real, "value": real}
                                        for real in self.realizations(self.ensembles[0])
                                    ],
                                    value=self.realizations(self.ensembles[0])[0],
                                    id=real_id,
                                    clearable=False,
                                    persistence=True,
                                    persistence_type="session",
                                ),
                                html.Button(
                                    style={
                                        "fontSize": "2rem",
                                        "paddingLeft": "5px",
                                        "paddingRight": "5px",
                                    },
                                    id=real_prev_id,
                                    children="⬅",
                                ),
                                html.Button(
                                    style={
                                        "fontSize": "2rem",
                                        "paddingLeft": "5px",
                                        "paddingRight": "5px",
                                    },
                                    id=real_next_id,
                                    children="➡",
                                ),
                            ],
                        ),
                    ]
                ),
            ]
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            id=self.uuid("layout"),
            children=[
                wcc.FlexBox(
                    style={"fontSize": "1rem"},
                    children=[
                        html.Div(
                            id=self.uuid("settings-view1"),
                            style={"margin": "10px", "flex": 4},
                            children=[
                                self.selector.layout,
                                self.ensemble_layout(
                                    ensemble_id=self.uuid("ensemble"),
                                    ens_prev_id=self.uuid("ensemble-prev"),
                                    ens_next_id=self.uuid("ensemble-next"),
                                    real_id=self.uuid("realization"),
                                    real_prev_id=self.uuid("realization-prev"),
                                    real_next_id=self.uuid("realization-next"),
                                ),
                            ],
                        ),
                        html.Div(
                            style={"margin": "10px", "flex": 4},
                            id=self.uuid("settings-view2"),
                            children=[
                                self.selector2.layout,
                                self.ensemble_layout(
                                    ensemble_id=self.uuid("ensemble2"),
                                    ens_prev_id=self.uuid("ensemble2-prev"),
                                    ens_next_id=self.uuid("ensemble2-next"),
                                    real_id=self.uuid("realization2"),
                                    real_prev_id=self.uuid("realization2-prev"),
                                    real_next_id=self.uuid("realization2-next"),
                                ),
                            ],
                        ),
                        html.Div(
                            style={"margin": "10px", "flex": 4},
                            id=self.uuid("settings-view3"),
                            children=[
                                html.Label("Calculation"),
                                html.Div(
                                    dcc.Dropdown(
                                        id=self.uuid("calculation"),
                                        value="Difference",
                                        clearable=False,
                                        options=[
                                            {"label": i, "value": i}
                                            for i in [
                                                "Difference",
                                                "Sum",
                                                "Product",
                                                "Quotient",
                                            ]
                                        ],
                                        persistence=True,
                                        persistence_type="session",
                                    )
                                ),
                                wcc.FlexBox(
                                    children=[
                                        html.Div(
                                            style={"width": "20%", "flex": 1},
                                            children=[
                                                html.Label("Truncate Min"),
                                                dcc.Input(
                                                    debounce=True,
                                                    type="number",
                                                    id=self.uuid("truncate-diff-min"),
                                                    persistence=True,
                                                    persistence_type="session",
                                                ),
                                            ],
                                        ),
                                        html.Div(
                                            style={"width": "20%", "flex": 1},
                                            children=[
                                                html.Label("Truncate Max"),
                                                dcc.Input(
                                                    debounce=True,
                                                    type="number",
                                                    id=self.uuid("truncate-diff-max"),
                                                    persistence=True,
                                                    persistence_type="session",
                                                ),
                                            ],
                                        ),
                                        html.Label(
                                            style={"fontSize": "1.2rem"},
                                            id=self.uuid("map3-label"),
                                        ),
                                    ]
                                ),
                            ],
                        ),
                    ],
                ),
                wcc.FlexBox(
                    style={"fontSize": "1rem"},
                    children=[
                        html.Div(
                            style={
                                "height": self.map_height,
                                "margin": "10px",
                                "flex": 4,
                            },
                            children=[
                                LeafletMap(
                                    syncedMaps=[self.uuid("map2"), self.uuid("map3")],
                                    id=self.uuid("map"),
                                    layers=[],
                                    unitScale={},
                                    autoScaleMap=True,
                                    minZoom=-5,
                                    updateMode="update",
                                    mouseCoords={"position": "bottomright"},
                                    colorBar={"position": "bottomleft"},
                                ),
                                html.Div(
                                    children=[
                                        dcc.RadioItems(
                                            id=self.uuid("hillshade"),
                                            labelStyle={
                                                "display": "inline-block",
                                                "text-align": "justify",
                                            },
                                            options=[
                                                {
                                                    "value": None,
                                                    "label": "No hillshading",
                                                },
                                                {
                                                    "value": "soft-hillshading",
                                                    "label": "Soft hillshading",
                                                },
                                                {
                                                    "value": "hillshading",
                                                    "label": "Hillshading",
                                                },
                                                {
                                                    "value": "hillshading_shadows",
                                                    "label": "Hillshading with shadows",
                                                },
                                            ],
                                            value=None,
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                    ]
                                ),
                            ],
                        ),
                        html.Div(
                            style={"margin": "10px", "flex": 4},
                            children=[
                                LeafletMap(
                                    syncedMaps=[self.uuid("map"), self.uuid("map3")],
                                    id=self.uuid("map2"),
                                    layers=[],
                                    unitScale={},
                                    autoScaleMap=True,
                                    minZoom=-5,
                                    updateMode="update",
                                    mouseCoords={"position": "bottomright"},
                                    colorBar={"position": "bottomleft"},
                                ),
                                html.Div(
                                    children=[
                                        dcc.RadioItems(
                                            id=self.uuid("hillshade2"),
                                            labelStyle={
                                                "display": "inline-block",
                                                "text-align": "justify",
                                            },
                                            options=[
                                                {
                                                    "value": None,
                                                    "label": "No hillshading",
                                                },
                                                {
                                                    "value": "soft-hillshading",
                                                    "label": "Soft hillshading",
                                                },
                                                {
                                                    "value": "hillshading",
                                                    "label": "Hillshading",
                                                },
                                                {
                                                    "value": "hillshading_shadows",
                                                    "label": "Hillshading with shadows",
                                                },
                                            ],
                                            value=None,
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                    ]
                                ),
                            ],
                        ),
                        html.Div(
                            style={"margin": "10px", "flex": 4},
                            children=[
                                LeafletMap(
                                    syncedMaps=[self.uuid("map"), self.uuid("map2")],
                                    id=self.uuid("map3"),
                                    layers=[],
                                    unitScale={},
                                    autoScaleMap=True,
                                    minZoom=-5,
                                    updateMode="update",
                                    mouseCoords={"position": "bottomright"},
                                    colorBar={"position": "bottomleft"},
                                ),
                                html.Div(
                                    children=[
                                        dcc.RadioItems(
                                            id=self.uuid("hillshade3"),
                                            labelStyle={
                                                "display": "inline-block",
                                                "text-align": "justify",
                                            },
                                            options=[
                                                {
                                                    "value": None,
                                                    "label": "No hillshading",
                                                },
                                                {
                                                    "value": "soft-hillshading",
                                                    "label": "Soft hillshading",
                                                },
                                                {
                                                    "value": "hillshading",
                                                    "label": "Hillshading",
                                                },
                                                {
                                                    "value": "hillshading_shadows",
                                                    "label": "Hillshading with shadows",
                                                },
                                            ],
                                            value=None,
                                            persistence=True,
                                            persistence_type="session",
                                        ),
                                    ]
                                ),
                            ],
                        ),
                        dcc.Store(
                            id=self.uuid("attribute-settings"),
                            data=json.dumps(self.attribute_settings),
                            storage_type="session",
                        ),
                    ],
                ),
            ],
        )

    def get_real_runpath(self, data: dict, ensemble: str, real: str) -> Path:
        filename = make_fmu_filename(data)
        runpath = Path(
            self.ens_df.loc[
                (self.ens_df["ENSEMBLE"] == ensemble) & (self.ens_df["REAL"] == real)
            ]["RUNPATH"].unique()[0]
        )

        return get_path(runpath / "share" / "results" / "maps" / f"{filename}.gri")

    def get_ens_runpath(self, data: dict, ensemble: str) -> List[Path]:
        filename = make_fmu_filename(data)
        runpaths = self.ens_df.loc[(self.ens_df["ENSEMBLE"] == ensemble)][
            "RUNPATH"
        ].unique()
        return [
            Path(runpath) / "share" / "results" / "maps" / f"{filename}.gri"
            for runpath in runpaths
        ]

    def set_callbacks(self, app: dash.Dash) -> None:
        @app.callback(
            [
                Output(self.uuid("map"), "layers"),
                Output(self.uuid("map2"), "layers"),
                Output(self.uuid("map3"), "layers"),
                Output(self.uuid("map3-label"), "children"),
            ],
            [
                Input(self.selector.storage_id, "data"),
                Input(self.uuid("ensemble"), "value"),
                Input(self.uuid("realization"), "value"),
                Input(self.selector2.storage_id, "data"),
                Input(self.uuid("ensemble2"), "value"),
                Input(self.uuid("realization2"), "value"),
                Input(self.uuid("calculation"), "value"),
                Input(self.uuid("attribute-settings"), "data"),
                Input(self.uuid("truncate-diff-min"), "value"),
                Input(self.uuid("truncate-diff-max"), "value"),
                Input(self.uuid("hillshade"), "value"),
                Input(self.uuid("hillshade2"), "value"),
                Input(self.uuid("hillshade3"), "value"),
            ],
        )
        # pylint: disable=too-many-arguments, too-many-locals
        def _set_base_layer(
            stored_selector_data: str,
            ensemble: str,
            real: str,
            stored_selector2_data: str,
            ensemble2: str,
            real2: str,
            calculation: str,
            stored_attribute_settings: str,
            diff_min: Union[int, float, None],
            diff_max: Union[int, float, None],
            hillshade: Union[str, None],
            hillshade2: Union[str, None],
            hillshade3: Union[str, None],
        ) -> Tuple[List[dict], List[dict], List[dict], str]:
            if not stored_selector_data or not stored_selector2_data:
                raise PreventUpdate

            # TODO(Sigurd)
            # These two are presumably of type dict, but the type depends on the actual python
            # objects that get serialized inside SurfaceSelector.
            # Should deserialization and validation be delegated to SurfaceSelector?
            # Note that according to the doc, it seems that dcc.Store actualy does the
            # serialization/deserialization for us!
            # Should be refactored
            data: dict = json.loads(stored_selector_data)
            data2: dict = json.loads(stored_selector2_data)
            if not isinstance(data, dict) or not isinstance(data2, dict):
                raise TypeError("Selector data payload must be of type dict")

            attribute_settings: dict = json.loads(stored_attribute_settings)
            if not isinstance(attribute_settings, dict):
                raise TypeError("Expected stored attribute_settings to be of type dict")

            if real in ["Mean", "StdDev", "Min", "Max"]:
                surface = calculate_surface(self.get_ens_runpath(data, ensemble), real)

            else:
                surface = load_surface(self.get_real_runpath(data, ensemble, real))
            if real2 in ["Mean", "StdDev", "Min", "Max"]:
                surface2 = calculate_surface(
                    self.get_ens_runpath(data2, ensemble2), real2
                )

            else:
                surface2 = load_surface(self.get_real_runpath(data2, ensemble2, real2))

            surface_layers: List[dict] = [
                make_surface_layer(
                    surface,
                    name="surface",
                    color=attribute_settings.get(data["attr"], {}).get("color"),
                    shader_type=hillshade,
                    min_val=attribute_settings.get(data["attr"], {}).get("min", None),
                    max_val=attribute_settings.get(data["attr"], {}).get("max", None),
                    unit=attribute_settings.get(data["attr"], {}).get("unit", " "),
                )
            ]
            surface_layers2: List[dict] = [
                make_surface_layer(
                    surface2,
                    name="surface2",
                    color=attribute_settings.get(data2["attr"], {}).get("color"),
                    shader_type=hillshade2,
                    min_val=attribute_settings.get(data2["attr"], {}).get("min", None),
                    max_val=attribute_settings.get(data2["attr"], {}).get("max", None),
                    unit=attribute_settings.get(data2["attr"], {}).get("unit", " "),
                )
            ]

            try:
                surface3 = calculate_surface_difference(surface, surface2, calculation)
                if diff_min is not None:
                    surface3.values[surface3.values <= diff_min] = diff_min
                if diff_max is not None:
                    surface3.values[surface3.values >= diff_max] = diff_max
                diff_layers: List[dict] = []
                diff_layers.append(
                    make_surface_layer(
                        surface3,
                        name="surface3",
                        color=attribute_settings.get(data["attr"], {}).get("color"),
                        shader_type=hillshade3,
                    )
                )
                error_label = ""
            except ValueError:
                diff_layers = []
                error_label = (
                    "Cannot calculate because the surfaces have different geometries"
                )

            if self.well_layer:
                surface_layers.append(self.well_layer)
                surface_layers2.append(self.well_layer)
                diff_layers.append(self.well_layer)
            return (surface_layers, surface_layers2, diff_layers, error_label)

        def _update_from_btn(
            _n_prev: int, _n_next: int, current_value: str, options: List[dict]
        ) -> str:
            """Updates dropdown value if previous/next btn is clicked"""
            option_values: List[str] = [opt["value"] for opt in options]
            ctx = dash.callback_context.triggered
            if not ctx or current_value is None:
                raise PreventUpdate
            if not ctx[0]["value"]:
                return current_value
            callback = ctx[0]["prop_id"]
            if "-prev" in callback:
                return prev_value(current_value, option_values)
            if "-next" in callback:
                return next_value(current_value, option_values)
            return current_value

        for btn_name in ["ensemble", "realization", "ensemble2", "realization2"]:
            app.callback(
                Output(self.uuid(f"{btn_name}"), "value"),
                [
                    Input(self.uuid(f"{btn_name}-prev"), "n_clicks"),
                    Input(self.uuid(f"{btn_name}-next"), "n_clicks"),
                ],
                [
                    State(self.uuid(f"{btn_name}"), "value"),
                    State(self.uuid(f"{btn_name}"), "options"),
                ],
            )(_update_from_btn)

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        store_functions: List[Tuple[Callable, list]] = [
            (
                find_surfaces,
                [
                    {
                        "ensemble_paths": self.ens_paths,
                        "suffix": "*.gri",
                        "delimiter": "--",
                    }
                ],
            )
        ]

        filenames = []
        # Generate all file names
        for attr, values in self.surfaceconfig.items():
            for name in values["names"]:
                for date in values["dates"]:
                    filename = f"{name}--{attr}"
                    if date is not None:
                        filename += f"--{date}"
                    filename += ".gri"
                    filenames.append(filename)

        # Copy all realization files
        for runpath in self.ens_df["RUNPATH"].unique():
            for filename in filenames:
                path = Path(runpath) / "share" / "results" / "maps" / filename
                if path.exists():
                    store_functions.append((get_path, [{"path": path}]))

        # Calculate and store statistics
        for _, ens_df in self.ens_df.groupby("ENSEMBLE"):
            runpaths = list(ens_df["RUNPATH"].unique())
            for filename in filenames:
                paths = [
                    Path(runpath) / "share" / "results" / "maps" / filename
                    for runpath in runpaths
                ]
                for statistic in ["Mean", "StdDev", "Min", "Max"]:
                    store_functions.append(
                        (save_surface, [{"fns": paths, "statistic": statistic}])
                    )
        if self.wellfolder is not None:
            store_functions.append(
                (find_files, [{"folder": self.wellfolder, "suffix": self.wellsuffix}])
            )
        if self.wellfiles is not None:
            store_functions.extend(
                [(get_path, [{"path": fn}]) for fn in self.wellfiles]
            )
        store_functions.append(
            (
                get_realizations,
                [
                    {
                        "ensemble_paths": self.ens_paths,
                        "ensemble_set_name": "EnsembleSet",
                    }
                ],
            )
        )
        return store_functions


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def calculate_surface(fns: List[str], statistic: str) -> xtgeo.RegularSurface:
    return surface_from_json(json.load(save_surface(fns, statistic)))


@webvizstore
def save_surface(fns: List[str], statistic: str) -> io.BytesIO:
    surfaces = xtgeo.Surfaces(fns)
    if len(surfaces.surfaces) == 0:
        surface = xtgeo.RegularSurface()
    elif statistic in ["Mean", "StdDev", "Min", "Max", "P10", "P90"]:
        # Suppress numpy warnings when surfaces have undefined z-values
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "All-NaN slice encountered")
            warnings.filterwarnings("ignore", "Mean of empty slice")
            warnings.filterwarnings("ignore", "Degrees of freedom <= 0 for slice")
            surface = calculate_statistic(surfaces, statistic)
    else:
        surface = xtgeo.RegularSurface()
    return io.BytesIO(surface_to_json(surface).encode())


# pylint: disable=too-many-return-statements
def calculate_statistic(
    surfaces: xtgeo.Surfaces, statistic: str
) -> xtgeo.RegularSurface:
    if statistic == "Mean":
        return surfaces.apply(np.nanmean, axis=0)
    if statistic == "StdDev":
        return surfaces.apply(np.nanstd, axis=0)
    if statistic == "Min":
        return surfaces.apply(np.nanmin, axis=0)
    if statistic == "Max":
        return surfaces.apply(np.nanmax, axis=0)
    if statistic == "P10":
        return surfaces.apply(np.nanpercentile, 10, axis=0)
    if statistic == "P90":
        return surfaces.apply(np.nanpercentile, 90, axis=0)
    return xtgeo.RegularSurface()


def surface_to_json(surface: xtgeo.RegularSurface) -> str:
    return json.dumps(
        {
            "ncol": surface.ncol,
            "nrow": surface.nrow,
            "xori": surface.xori,
            "yori": surface.yori,
            "rotation": surface.rotation,
            "xinc": surface.xinc,
            "yinc": surface.yinc,
            "values": surface.values.copy().filled(np.nan).tolist(),
        }
    )


def surface_from_json(surfaceobj: dict) -> xtgeo.RegularSurface:
    # See https://github.com/equinor/xtgeo/issues/405
    surface = xtgeo.RegularSurface(**surfaceobj)
    surface.values = np.array(surfaceobj["values"])
    return surface


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_surfaces(fns: List[str]) -> xtgeo.Surfaces:
    return xtgeo.Surfaces(fns)


@webvizstore
def get_path(path: Path) -> Path:
    return Path(path)


def prev_value(current_value: str, options: List[str]) -> str:
    try:
        index = options.index(current_value)
        return options[max(0, index - 1)]
    except ValueError:
        return current_value


def next_value(current_value: str, options: List[str]) -> str:
    try:
        index = options.index(current_value)
        return options[min(len(options) - 1, index + 1)]

    except ValueError:
        return current_value


def surfacedf_to_dict(df: pd.DataFrame) -> dict:
    return {
        attr: {
            "names": list(dframe["name"].unique()),
            "dates": list(dframe["date"].unique())
            if "date" in dframe.columns
            else None,
        }
        for attr, dframe in df.groupby("attribute")
    }


@webvizstore
def find_files(folder: Path, suffix: str) -> io.BytesIO:
    return io.BytesIO(
        json.dumps(
            sorted([str(filename) for filename in folder.glob(f"*{suffix}")])
        ).encode()
    )


def make_fmu_filename(data: dict) -> str:
    filename = f"{data['name']}--{data['attr']}"
    if data["date"] is not None:
        filename += f"--{data['date']}"
    return filename


def calculate_surface_difference(
    surface: xtgeo.RegularSurface,
    surface2: xtgeo.RegularSurface,
    calculation: str = "Difference",
) -> xtgeo.RegularSurface:
    surface3 = surface.copy()
    if calculation == "Difference":
        surface3.values = surface3.values - surface2.values
    elif calculation == "Sum":
        surface3.values = surface3.values + surface2.values
    elif calculation == "Product":
        surface3.values = surface3.values * surface2.values
    elif calculation == "Quotient":
        surface3.values = surface3.values / surface2.values
    return surface3



from dash.dependencies import Input, Output, State
import dash_table
import webviz_core_components as wcc


def parameter_qc_controller(parent, app):
    @app.callback(
        Output(parent.uuid("property-qc-wrapper"), "children"),
        Input({"id": parent.uuid("ensemble-selector"), "tab": "qc"}, "value"),
        Input({"id": parent.uuid("delta-ensemble-selector"), "tab": "qc"}, "value"),
        Input(
            {"id": parent.uuid("filter-parameter"), "tab": "qc"},
            "value",
        ),
        Input(parent.uuid("property-qc-plot-type"), "value"),
    )
    def _update_bars(ensemble, delta_ensemble, parameters, plot_type):
        """Callback to switch visualization between table and distribution plots"""
        parameters = parameters if isinstance(parameters, list) else [parameters]
        ensembles = [ensemble, delta_ensemble]

        if plot_type == "table":
            columns, dframe = parent.pmodel.make_statistics_table(
                ensembles=ensembles, parameters=parameters
            )
            return dash_table.DataTable(
                style_table={
                    "height": "75vh",
                    "overflow": "auto",
                    "fontSize": "1.5rem",
                },
                columns=columns,
                data=dframe,
                sort_action="native",
                filter_action="native",
            )
        return wcc.Graph(
            id=parent.uuid("property-qc-graph"),
            config={"displayModeBar": False},
            style={"height": "75vh"},
            figure=parent.pmodel.make_grouped_plot(
                ensembles=ensembles,
                parameters=parameters,
                plot_type=plot_type,
            ),
        )

    @app.callback(
        Output(
            {"id": parent.uuid("filter-parameter"), "tab": "qc"},
            "options",
        ),
        Output(
            {"id": parent.uuid("filter-parameter"), "tab": "qc"},
            "value",
        ),
        Input(parent.uuid("delta-sort"), "value"),
        Input({"id": parent.uuid("ensemble-selector"), "tab": "qc"}, "value"),
        Input({"id": parent.uuid("delta-ensemble-selector"), "tab": "qc"}, "value"),
        State(
            {"id": parent.uuid("filter-parameter"), "tab": "qc"},
            "value",
        ),
    )
    def _update_parameters(sortby, ensemble, delta_ensemble, current_params):
        """Callback to sort parameters based on selection"""
        parent.pmodel.sort_parameters(
            ensemble=ensemble,
            delta_ensemble=delta_ensemble,
            sortby=sortby,
        )
        return [
            {"label": i, "value": i} for i in parent.pmodel.parameters
        ], current_params



import sys
import subprocess  # nosec
from pathlib import Path

from dash.testing.composite import DashComposite
from selenium.webdriver.support.wait import WebDriverWait
from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
)


def test_full_example(
    testdata_folder: Path, dash_duo: DashComposite, tmp_path: Path
) -> None:

    # https://github.com/plotly/dash/issues/1164:
    # We are accessing a private member here which seems to be necessary due to the
    # aforementioned issue. Ignore the respective pylint warning.
    # pylint: disable=protected-access
    dash_duo._wd_wait = WebDriverWait(
        dash_duo.driver,
        timeout=10,
        ignored_exceptions=(NoSuchElementException, StaleElementReferenceException),
    )
    # pylint: enable=protected-access

    # Build a portable webviz from config file
    appdir = tmp_path / "app"
    subprocess.call(  # nosec
        ["webviz", "build", "webviz-raw-data.yml", "--portable", appdir],
        cwd=testdata_folder / "webviz_examples",
    )
    # Remove Talisman
    file_name = appdir / "webviz_app.py"
    with open(file_name, "r") as file:
        lines = file.readlines()
    with open(file_name, "w") as file:
        for line in lines:
            if not line.strip("\n").startswith("Talisman"):
                file.write(line)
    # Import generated app
    sys.path.append(str(appdir))

    # webviz_app was just created, temporarily ignore the import-outside-toplevel warning
    # and the import-error.
    # pylint: disable=import-outside-toplevel
    # pylint: disable=import-error
    from webviz_app import app

    # pylint: enable=import-outside-toplevel
    # pylint: enable=import-error

    # Start and test app
    dash_duo.start_server(app)
    for page in [
        "inplacevolumesonebyone",
        "reservoirsimulationtimeseriesonebyone",
        "inplacevolumes",
        "parameterdistribution",
        "parametercorrelation",
        "reservoirsimulationtimeseries",
    ]:
        dash_duo.wait_for_element(f"#{page}").click()
        logs = [
            log
            for log in dash_duo.get_logs()
            if "TypeError: Cannot read property 'hardwareConcurrency' of undefined"
            not in log["message"]
        ]

        if logs != []:
            raise AssertionError(page, logs)



from typing import List, Tuple

import webviz_core_components as wcc
import pandas as pd
import numpy as np


def update_crossplot(df: pd.DataFrame, sizeby: str, colorby: str) -> List[wcc.Graph]:

    sim_range = find_sim_range(df)
    sizeref, cmin, cmax = size_color_settings(df, sizeby, colorby)

    figures = []

    for _ens, ensdf in df.groupby("ENSEMBLE"):

        dframe = (
            ensdf.groupby(["WELL", "DATE", "ZONE", "TVD"]).mean().reset_index().copy()
        )
        trace = {
            "x": dframe["OBSERVED"],
            "y": dframe["SIMULATED"],
            "type": "scatter",
            "mode": "markers",
            "hovertext": [
                f"Well: {well}"
                f"<br>Zone: {zone}"
                f"<br>Pressure observation: {obs:.2f}"
                f"<br>Mean simulated pressure: {pressure:.2f}"
                f"<br>Mean misfit: {misfit:.2f}"
                f"<br>Stddev pressure: {stddev:.2f}"
                for well, zone, obs, stddev, misfit, pressure in zip(
                    dframe["WELL"],
                    dframe["ZONE"],
                    dframe["OBSERVED"],
                    dframe["STDDEV"],
                    dframe["DIFF"],
                    dframe["SIMULATED"],
                )
            ],
            "hoverinfo": "text",
            "marker": {
                "size": dframe[sizeby],
                "sizeref": 2.0 * sizeref / (30.0 ** 2),
                "sizemode": "area",
                "sizemin": 6,
                "color": dframe[colorby],
                "cmin": cmin,
                "cmax": cmax,
                "colorscale": [[0, "#2584DE"], [1, "#E50000"]],
                "colorbar": {"x": 1.05},
                "showscale": True,
            },
        }

        layout = {
            "height": 400,
            "title": {
                "text": _ens,
                "y": 0.95,
                "x": 0.15,
                "xanchor": "center",
                "yanchor": "top",
            },
            "margin": {"l": 100, "r": 0, "b": 50, "t": 30},
            "showlegend": False,
            "xaxis": {
                "range": sim_range,
                "title": "Pressure Observation",
                "showticklabels": True,
            },
            "yaxis": {"range": sim_range, "title": "Simulated mean pressure"},
            "shapes": [
                {
                    "type": "line",
                    "x0": sim_range[0],
                    "y0": sim_range[0],
                    "x1": sim_range[1],
                    "y1": sim_range[1],
                    "line": {
                        "color": "#007079",
                        "width": 2,
                    },
                }
            ],
        }

        figures.append(wcc.Graph(figure={"data": [trace], "layout": layout}))
    return figures


def size_color_settings(
    df: pd.DataFrame, sizeby: str, colorby: str
) -> Tuple[np.float64, np.float64, np.float64]:

    df = df.groupby(["WELL", "DATE", "ZONE", "TVD", "ENSEMBLE"]).mean().reset_index()

    sizeref = df[sizeby].quantile(0.9)
    cmin = df[colorby].min()
    cmax = df[colorby].quantile(0.9)

    return sizeref, cmin, cmax


def find_sim_range(df: pd.DataFrame) -> List[np.float64]:

    df = df.groupby(["WELL", "DATE", "ZONE", "TVD", "ENSEMBLE"]).mean().reset_index()

    max_sim = (
        df["SIMULATED"].max()
        if df["SIMULATED"].max() > df["OBSERVED"].max()
        else df["OBSERVED"].max()
    )
    min_sim = (
        df["SIMULATED"].min()
        if df["SIMULATED"].min() < df["OBSERVED"].min()
        else df["OBSERVED"].min()
    )

    axis_extend = (max_sim - min_sim) * 0.1

    return [min_sim - axis_extend, max_sim + axis_extend]



import mock
import dash
import pandas as pd
from webviz_config.common_cache import CACHE
from webviz_config.themes import default_theme

# pylint: disable=no-name-in-module
from webviz_config.plugins import (
    ParameterCorrelation,
)

# pylint: enable=no-name-in-module

# mocked functions
GET_PARAMETERS = "webviz_subsurface.plugins._parameter_correlation.get_parameters"


def test_parameter_corr(dash_duo: dash.testing.composite.DashComposite) -> None:

    app = dash.Dash(__name__)
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    app.webviz_settings = {
        "shared_settings": {"scratch_ensembles": {"iter-0": ""}},
        "theme": default_theme,
    }
    ensembles = ["iter-0"]

    with mock.patch(GET_PARAMETERS) as mock_parameters:
        mock_parameters.return_value = pd.read_csv("tests/data/parameters.csv")

        parameter_correlation = ParameterCorrelation(app, ensembles)

        app.layout = parameter_correlation.layout
        dash_duo.start_server(app)

        my_component = dash_duo.find_element(
            f"#{parameter_correlation.ids('ensemble-all')}"
        )

        if not my_component.text.startswith("iter-0"):
            raise AssertionError()



from typing import Tuple


def hex_to_rgb(hex_string: str, opacity: float = 1) -> str:
    """Converts a hex color to rgb"""
    hex_string = hex_string.lstrip("#")
    hlen = len(hex_string)
    rgb = [int(hex_string[i : i + hlen // 3], 16) for i in range(0, hlen, hlen // 3)]
    rgb.append(opacity)
    return f"rgba{tuple(rgb)}"


def rgb_to_hex(color):
    """Converts a rgb color to hex"""
    color = color.strip("rgb()")
    color = color.split(",")
    return "#%02x%02x%02x" % (int(color[0]), int(color[1]), int(color[2]))


def find_intermediate_color(
    lowcolor: str, highcolor: str, intermed: float, colortype: str = "tuple"
) -> str:
    """
    Returns the color at a given distance between two colors
    This function takes two color tuples, where each element is between 0
    and 1, along with a value 0 < intermed < 1 and returns a color that is
    intermed-percent from lowcolor to highcolor. If colortype is set to 'rgb',
    the function will automatically convert the rgb type to a tuple, find the
    intermediate color and return it as an rgb color.
    """

    if colortype == "rgba":
        # convert to tuple color, eg. (1, 0.45, 0.7)
        lowcolor = unlabel_rgba(lowcolor)
        highcolor = unlabel_rgba(highcolor)

    diff_0 = float(highcolor[0] - lowcolor[0])
    diff_1 = float(highcolor[1] - lowcolor[1])
    diff_2 = float(highcolor[2] - lowcolor[2])
    diff_3 = float(highcolor[3] - lowcolor[3])

    inter_med_tuple = (
        lowcolor[0] + intermed * diff_0,
        lowcolor[1] + intermed * diff_1,
        lowcolor[2] + intermed * diff_2,
        lowcolor[3] + intermed * diff_3,
    )

    if colortype == "rgba":
        # back to an rgba string, e.g. rgba(30, 20, 10)
        inter_med_rgba = label_rgba(inter_med_tuple)
        return inter_med_rgba

    return inter_med_tuple


def label_rgba(colors: str) -> str:
    """
    Takes tuple (a, b, c, d) and returns an rgba color 'rgba(a, b, c, d)'
    """
    return "rgba(%s, %s, %s, %s)" % (colors[0], colors[1], colors[2], colors[3])


def unlabel_rgba(colors: str) -> Tuple[float, float, float, float]:
    """
    Takes rgba color(s) 'rgba(a, b, c, d)' and returns tuple(s) (a, b, c, d)
    This function takes either an 'rgba(a, b, c, d)' color or a list of
    such colors and returns the color tuples in tuple(s) (a, b, c, d)
    """
    str_vals = ""
    for index, _col in enumerate(colors):
        try:
            float(colors[index])
            str_vals = str_vals + colors[index]
        except ValueError:
            if colors[index] == "," or colors[index] == ".":
                str_vals = str_vals + colors[index]

    str_vals = str_vals + ","
    numbers = []
    str_num = ""
    for char in str_vals:
        if char != ",":
            str_num = str_num + char
        else:
            numbers.append(float(str_num))
            str_num = ""
    return tuple(numbers)
